/*
 * SuperY.c
 *
 *  Created on: Jun 16, 2011
 *      Author: Brenda Javornik (e-mail: brendaj (at) cora.nwra.com)
 */

#include <string.h>
#include <complex.h>
#include "errorHandler.h"
#include "MatrixFunctionErrorMessages.h"
#include "utilities.h"
#include "matrix.h"
#include "MapDepthMatrix.h"
#include "SuperY.h"

/** Initialize super vector Y.
 * @param[in,out] y Pointer to SuperVectorY with allocated space; return initialized vector.
 * @param[in] m Defining kernel matrix in wave space
 */
	void InitSuperVectorY(SuperVectorY *y, MAP_DEPTH_MATRIX m) {
	y->Nx = m.nX;
	y->Ny = (m.nY - 1) * 2;
	y->Nz = m.nZ;
	y->Nm = 1; // because Y contains averaging kernels and they are independent of maps
	y->Nv = m.nV;
	y->noiseBase = y->Nx * y->Ny * y->Nz * y->Nv;
	y->RC = y->Nx * y->Ny;
}



/** Convert super vector Y (which is double) to double complex.
 * @param[in] rData Pointer to double array for real data; must be allocated.
 * @param[in] nelements Number of elements in rData.
 * @param[in,out] cData Pointer to double complex data; must be allocated.  Complex
 * form of real data returned in this array.
 */
void RealToDoubleComplex(double *rData, int nelements, double complex *cData) {

	if (rData == NULL)
		reportError(SuperYRDataNull, "RealToDoubleComplex", FATAL);
	if (cData == NULL)
		reportError(SuperYCDataNull, "RealToDoubleComplex", FATAL);
	int i;
	double *dp;
	dp = rData;
	double complex *dcp;
	dcp = cData;
	for (i = 0; i < nelements; i++) {
		*dcp++ = *dp++ + 0.0 * I;
	}
}

/** Convert from double complex to double
 * @param[in] cData Pointer to array of double complex data; cannot be empty.
 * @param[in] nelements Number of elements in cData.
 * @param[in,out] rData Pointer to array of doubles; must be allocated.  Real
 * form of complex data returned in this array.
 * @return FALSE if error; TRUE if successful
 */
BOOLEAN DoubleComplexToReal(double complex *cData, int nelements, double *rData) {

	BOOLEAN conversionOk = TRUE;
	if (rData == NULL)
		reportError(SuperYRDataNull, "DoubleComplexToReal", FATAL);
	if (cData == NULL)
		reportError(SuperYCDataNull, "DoubleComplexToReal", FATAL);
	int i;
	double *dp;
	dp = rData;
	double complex *dcp;
	dcp = cData;
	for (i = 0; i < nelements; i++) {
		if (cimag(*dcp) != 0.0) {
			if (conversionOk) {
				char msg[MAX_ERROR_MSG_LENGTH];
				snprintf(msg, MAX_ERROR_MSG_LENGTH, SuperYCannotConvertCToI, i, creal(*dcp), cimag(*dcp));
				reportError(msg, "DoubleComplexToReal", ERROR);
				conversionOk = FALSE;
			}
		}
		*dp++ = creal(*dcp++); // ignoring imaginary part
	}
	return conversionOk;
}

/**
 * Reconstitute averaging kernel and noise matrices from packed super vector Y.
 * @param[in] superY is in column-major order.
 * y must be in real space, NOT wave space.
 * @param[in,out] AvgKern is in real space. Data is overwritten. Send already allocated matrix.
 * @param[in,out] noise is in real space. Data is overwritten. Send already allocated matrix.
 * @param[in] waveSpace If true, then averaging kernel and noise are in wave space; NOT IMPLEMENTED!
 * otherwise, they are in real space.
 */
void unpackSuperY(const SuperVectorY *superY, MAP_DEPTH_MATRIX *AvgKern, MAP_DEPTH_MATRIX *noise, BOOLEAN waveSpace) {

	//MAP_DEPTH_MATRIX temp;
	int Nx, NyOrig, Nz, Nv, Nt;

	Nx = AvgKern->nX;
	NyOrig = AvgKern->nY;
	Nz = AvgKern->nZ;
	Nt = AvgKern->nT;
	Nv = AvgKern->nV;

	// unpack averaging kernel first
	// pull from superY; superY is always in real space and double
	int nelements;
	nelements = Nx*NyOrig*Nz*Nt*Nv;
	RealToDoubleComplex(superY->yElements,  nelements, AvgKern->data);

	// pull noise from superY next

	// convert from complex to real
	nelements = noise->nX * noise->nY * noise->nZ * noise->nT * noise->nV;
	RealToDoubleComplex(&(superY->yElements[superY->noiseBase]), nelements, noise->data);
}

/**
 * Condense averaging kernel and noise matrices into packed super vector Y.
 * [AvgKern; noise] averaging kernel is at the top of Y, noise is in the bottom.
 * @param[in,out] AvgKern is in real space. Data is overwritten.
 * @param[in,out] noise is in real space. Data is overwritten.
 * @param[in] superY is in column-major order. ??  Send already allocated matrix.
 * y must be in real space, NOT wave space.
 * @param[in] realSpace If true, then averaging kernel and noise are in real space; NOT IMPLEMENTED!
 * otherwise, they are in wave space.
 * @return false if packing failed; true if successful.
 */
BOOLEAN packSuperY(const MAP_DEPTH_MATRIX *AvgKern, const MAP_DEPTH_MATRIX *noise, SuperVectorY *superY, BOOLEAN realSpace) {

	BOOLEAN packOk = TRUE;
	// pack averaging kernel first

	// convert from complex to real
	int nelementsAK;
	nelementsAK = AvgKern->nX * AvgKern->nY * AvgKern->nZ * AvgKern->nT * AvgKern->nV;

	int nelementsNoise;
	nelementsNoise = noise->nX * noise->nY * noise->nZ * noise->nT * noise->nV;

	if (nelementsAK + nelementsNoise > superY->yLength)
		reportError(SuperYAvgKNoiseExceed, "packSuperY", FATAL);

	if (!DoubleComplexToReal(AvgKern->data, nelementsAK, superY->yElements)) {
		reportError(SuperYAvgKIToReal,"packSuperY", ERROR);
		packOk = FALSE;
	}

	// pack noise second

	// convert from complex to real

	int noiseBaseIdx;
	noiseBaseIdx = superY->noiseBase;
	if (!DoubleComplexToReal(noise->data, nelementsNoise, &(superY->yElements[noiseBaseIdx]))) {
		reportError(SuperYNoiseIToReal,"packSuperY", ERROR);
		packOk = FALSE;
	}
	return packOk;

}
