/*
 * fftw.c
 *
 *  Created on: Apr 16, 2010
 *      Author: Brenda Javornik (e-mail: brendaj (at) cora.nwra.com)
 */

#include <string.h>
#include <stdio.h>
#include <math.h>
// include complex.h and fftw will use the native c complex datatype (double complex);
// then fftw_complex is defined to be the native complex type
#include <complex.h>
#include <fftw3.h>
#include "utilities.h"
#include "objectfits.h"
#include "matrix.h"
#include "errorHandler.h"
#include "MatrixFunctionErrorMessages.h"


	int reversePlanDone = FALSE;	///< fftw plans are reused as long as the size of input and output data remain the same.
	fftw_plan fplaninverse = NULL;	///< fftw reverse plan
	int RinX = 0; 					///< x dimension of input data for reverse plan
	int RinY = 0;					///< y dimension of input data for reverse plan

MATRIX fftwShift(MATRIX m);

/** fftw Performs a forward Fourier transform using the Fasted Fourier Transform in
 * the West (FFTW) library.
 * @param[in] function A fits2d object which contains data values for a function to transform.
 * @return A MATRIX object containing the transformed data.
 * @ingroup Matrix
 */
// use fftw_malloc to create input function data
MATRIX fftw(fits2d function) {
	double *in;
	fftw_complex *out;
	static int forwardPlanDone = FALSE;
	static fftw_plan fplanforward = NULL;
	static int inX = 0, inY = 0;
	double norm, normI;
	int x;

	// transform the data

	in = function.data;

	out = (fftw_complex *) fftw_malloc(function.xDim * (function.yDim / 2 + 1)
			* sizeof(fftw_complex));
	if (out == NULL)
		reportError(OutOfMemory, "fftw", FATAL);

	// make a new plan, if one doesn't exist or the dimensions have changed
	if ((!forwardPlanDone) || (inX != function.xDim) || (inY != function.yDim)) {
		if (fplanforward != NULL)
			fftw_destroy_plan(fplanforward);
		fplanforward = fftw_plan_dft_r2c_2d(function.xDim, function.yDim, in,
				out,
				FFTW_ESTIMATE);
		forwardPlanDone = TRUE;
		inX = function.xDim;
		inY = function.yDim;

		fplaninverse = fftw_plan_dft_c2r_2d(function.xDim, function.yDim, out, in, FFTW_ESTIMATE);
		reversePlanDone = TRUE;
		RinX = function.xDim;
		RinY = function.yDim;

		// The plan is associated with the out pointer;
		// if we change the pointer, then we may need to make a new plan or call fftw_execute_xxx instead
		// of fftw_execute();
	}

	// reuse the plan
	fftw_execute_dft_r2c(fplanforward, in, out);

	norm = 1.0 / ((2.0 * M_PI) * (2.0 * M_PI));
	normI = ((2.0 * M_PI) * (2.0 * M_PI)) / (function.xDim * function.xDim);

	//scale the data
	for (x=0; x<function.xDim * (function.yDim / 2 + 1); x++)
		out[x] = out[x] * norm;

	// convert out matrix to a MATRIX datatype
	MATRIX result;

	result = newComplexMatrix(function.xDim, (function.yDim / 2 + 1));
	td_free(result.data, function.xDim * (function.yDim / 2 + 1) * sizeof(double complex));
	result.data = out;
	return result;

}

/** fftwReverse Performs an inverse Fourier transform using the Fasted Fourier Transform in
 * the West (FFTW) library.
 * @param[in] function A MATRIX object which contains data values for a function to transform.
 * @param[in] shift A boolean value whether to shift the data values.  TRUE means shift the values.
 * FALSE means no shift.
 * @return A MATRIX object containing the transformed data.
 * @ingroup Matrix
 */
MATRIX fftwReverse(MATRIX function, int shift) {
	double *in;
	double complex *out;
	static
	double complex *outcopy = NULL;
	int fullSize;

	int x;

	fullSize = 2*(function.nCol-1);
	// transform the data

	// allocate double space for output of ifftw
	int inSize = function.nRow * fullSize * sizeof(double);
	in = (double*) fftw_malloc(inSize);
	if (in == NULL)
		reportError(OutOfMemory, "fftwReverse", FATAL);

	out = function.data;

	// get memory corruption here !!!!!!!! for fftw_plan_dft_c2r_2d
	// we are doing an out-of-place transform.
	// FFTW_PRESERVE_INPUT specifies that an out-of-place transform must not change its input array.
	// This is ordinarily the default, except for c2r and hc2r (i.e. complex-to-real) transforms for which
	// FFTW_DESTROY_INPUT is the default. In the latter cases, passing FFTW_PRESERVE_INPUT will attempt to
	// use algorithms that do not destroy the input, at the expense of worse performance;
	// for multi-dimensional c2r transforms, however, no input-preserving algorithms are implemented and
	// the planner will return NULL if one is requested.

	// make a new plan, if one doesn't exist or the dimensions have changed
	if ((!reversePlanDone) || (RinX != function.nRow) || (RinY != fullSize)) {
		if (fplaninverse != NULL)
			fftw_destroy_plan(fplaninverse);

		fplaninverse = fftw_plan_dft_c2r_2d(function.nRow, fullSize, outcopy, in, FFTW_ESTIMATE);
		reversePlanDone = TRUE;
		RinX = function.nRow;
		RinY = function.nCol;
	}

	// reuse the plan
	fftw_execute_dft_c2r(fplaninverse, out, in);

	// convert out matrix to a MATRIX datatype
	MATRIX result;
	double *p;

	result = newMatrix(function.nRow, 2*(function.nCol-1));
	td_free(result.data, result.nCol * result.nRow * sizeof(double));
	result.data = in;

	double norm = ((2.0 * M_PI) * (2.0 * M_PI));
	p = result.data;
	// scale the data
	for (x=0; x<result.nRow*result.nCol; x++) {
		p[x] =  p[x] / (result.nRow * result.nCol) * norm;
	}

	if (shift) {
		MATRIX shiftedResult = fftwShift(result);
		fftw_free(result.data);
		return shiftedResult;
	} else {
		return result;
	}
}

// shift data from this format
// 2 1
// 4 3
//
// to this format
// 3 4
// 1 2
//
/** fftwShift shifts data values after a reverse transform.  Because of the way FFTW
 * performs the Fourier transform, the data sometimes needs to be shifted.
 * @param[in] m A MATRIX containing data to shift.
 * @return A MATRIX containing shifted data.
 * @ingroup Matrix
 */
MATRIX fftwShift(MATRIX m) {

	MATRIX result;

	if (m.mtype != REAL) {
		reportError(FftwShiftNoComplexData, "", FATAL);
	} else {
		int i;
		int toR, toC, fromR, fromC;
		int halfRow, halfCol;
		double *p, *q;

		result = newMatrix(m.nRow, m.nCol);
		p = m.data;
		q = result.data;
		halfRow = m.nRow/2;
		halfCol = m.nCol/2;
		for (i=0; i<m.nRow/2; i++) {
			// memcpy(to, from, size)
			// copy quadrant 1
			toR = halfRow + i;
			toC = 0;
			fromR = 0 + i;
			fromC = halfCol;
			memcpy(&q[ROW_MAJOR_IDX(toR, toC, m.nCol)], &p[ROW_MAJOR_IDX(fromR, fromC, m.nCol)], halfCol * sizeof(double));
			// copy quadrant 2
			toR = halfRow + i;
			toC = halfCol;
			fromR = 0 + i;
			fromC = 0;
			memcpy(&q[ROW_MAJOR_IDX(toR, toC, m.nCol)], &p[ROW_MAJOR_IDX(fromR, fromC, m.nCol)], halfCol * sizeof(double));
			// copy quadrant 3
			toR = 0 + i;
			toC = 0;
			fromR = halfRow + i;
			fromC = halfCol;
			memcpy(&q[ROW_MAJOR_IDX(toR, toC, m.nCol)], &p[ROW_MAJOR_IDX(fromR, fromC, m.nCol)], halfCol * sizeof(double));
			// copy quadrant 4
			toR = 0 + i;
			toC = halfCol;
			fromR = halfRow + i;
			fromC = 0;
			memcpy(&q[ROW_MAJOR_IDX(toR, toC, m.nCol)], &p[ROW_MAJOR_IDX(fromR, fromC, m.nCol)], halfCol * sizeof(double));
		}
	}

	return result;
}

