/*
 * linearEqSolve.c
 *
 *  Created on: Apr 16, 2010
 *      Author: Brenda Javornik (e-mail: brendaj (at) cora.nwra.com)
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <complex.h>
#include "utilities.h"
#include "matrix.h"
#include "errorHandler.h"
#include "MatrixFunctionErrorMessages.h"
#include "ProblemDimension.h"

/** External LAPACK function */
void dgesv_(const int *N, const int *nrhs, double *A, const int *lda, int
	*ipiv, double *b, const int *ldb, int *info);
/** External LAPACK function */
void zgesv_(const int *N, const int *nrhs, double complex *A, const int *lda, int
	*ipiv, double complex *b, const int *ldb, int *info);

/** External LAPACK function */
void dgels_(const char *trans, const int *M, const int *N, const int *nrhs,
    double *A, const int *lda, double *b, const int *ldb, double *work,
    const int * lwork, int *info);

/** External LAPACK function */
void dgetrs_(const char *trans, const int *N, const int *nrhs,
		double *A, const int *lda, const int *ipiv, double *b, const int *ldb, int *info);
/** External LAPACK function */
void zgetrs_(const char *trans, const int *N, const int *nrhs,
		double complex *A, const int *lda, const int *ipiv, double complex *b, const int *ldb, int *info);

/** External LAPACK function */
void dgetrf_(const int *M, const int *N, double *A, const int *lda, int
	*ipiv, int *info);

/* Keep the LUDecomp and pivot values of the last call to linearEqSolve
 * because it can be reused when calculating the
 * Averaging Kernel.
 *
 */
static MATRIX LUDecomp;	///< Keep the LUDecomp values of the last call to linearEqSolve because it can be reused when calculating the Averaging Kernel.
static int *IPiv;		///< Keep the pivot values of the last call to linearEqSolve because it can be reused when calculating the Averaging Kernel.

/**
 * Solve Ax=B for x.
 * @param[in] A Matrix
 * @param[in] B Matrix
 * @return A Matrix containing the solution x.
 * @ingroup Matrix
 */
MATRIX
linearEqSolve(MATRIX A, MATRIX B)
{
    MATRIX result;

    int N; 	// The number of linear equations
    int nrhs;
    int lda;
    int* ipiv;
    int ldb;
    int info;

    // leading dimension = # of rows

    N = B.nRow;
    nrhs = B.nCol;
    lda = A.nRow;

    ipiv = (int *) td_malloc(N * sizeof(int));
	if (ipiv == NULL)
		reportError(OutOfMemory, "linearEqSolve.ipiv", FATAL);
    ldb = B.nRow;

    if (LUDecomp.data != NULL) {
    	freeMatrix(LUDecomp); // LUDecomp.data);
    }

    if (IPiv != NULL) {
    	td_free(IPiv, LUDecomp.nRow * sizeof(int));
    }

	if ((A.mtype == REAL) && (B.mtype == REAL)) {
		result = newMatrix(A.nCol, B.nCol);
		td_free(result.data, A.nCol * B.nCol * sizeof(double));

	    LUDecomp = newMatrix(A.nRow, A.nCol);
		memcpy(LUDecomp.data, A.data, (A.nRow * A.nCol * sizeof(double)));
		/*	test
	    td_free(LUDecomp.data, A.nRow * A.nCol * sizeof(double));
		// switch LUDecomp and VCopy to row major order
		LUDecomp = toColumnMajor2(A);
		*/
		double* VCopy;
		VCopy = (double *) td_malloc(B.nRow * B.nCol * sizeof(double));
		if (VCopy == NULL)
			reportError(OutOfMemory, "linearEqSolve.VCopy", FATAL);
		memcpy(VCopy, B.data, (B.nRow * B.nCol * sizeof(double)));

		// for a vector, row major order is the same as column major order
		// when reusing the LUDecomp remember to transpose the LUDecomp matrix
		if (B.nCol != 1)
			reportError(LinearEqSolveExpectedColVector, "", FATAL);
		dgesv_(&N, &nrhs, LUDecomp.data, &lda, ipiv, VCopy, &ldb, &info);
		result.data = VCopy;
	} else {
		// complex data

	    LUDecomp = newComplexMatrix(A.nRow, A.nCol);
	    td_free(LUDecomp.data, A.nRow * A.nCol * sizeof(double complex));
		// switch LUDecomp and VCopy to row major order
		LUDecomp = toColumnMajor2(A);
		MATRIX VCopy;
		VCopy = toColumnMajor2(B);
		zgesv_(&N, &nrhs, LUDecomp.data, &lda, ipiv, VCopy.data, &ldb, &info);
		// when reusing the LUDecomp do NOT transpose the LUDecomp matrix
		result = toRowMajor2(VCopy);

		freeMatrix(VCopy);
	}

    // save intermediate values
    IPiv = ipiv;

    if(info == 0) { /* succeed */
    	;
    }
    else {
    	char msg[MAX_ERROR_MSG_LENGTH];
    	if (info < 0) {
    		// if INFO = -i, the i-th argument had an illegal value
    		snprintf(msg, MAX_ERROR_MSG_LENGTH, "Argument %d had an illegal value", -info);
    	} else {
    		//from description of dgesv.f
    		//  if INFO = i, U(i,i) is exactly zero.  The factorization
    		//                has been completed, but the factor U is exactly
    		//                singular, so the solution could not be computed.
    		char msg1[] = "U(i,i) is exactly zero.  The factorization has been completed, but the factor U is exactly singular, so the solution could not be computed.";
    		snprintf(msg, MAX_ERROR_MSG_LENGTH, "For i = %d, %s", info, msg1);
    	}
    	reportError(DGESVFailed, msg, WARNING);
    	// recover by returning a NaN vector or matrix
    	setToNaNMatrix(result);
    }
	return result;
}

/** Same as above except, by changing nrhs to the number of columns in v, we can use this to solve many systems of
 * linear equations or invert a matrix.
 * Solves A * X = B for X
 * @param[in] A Matrix
 * @param[in] B Matrix
 * @return A Matrix containing the solution x.
 * @ingroup Matrix
 */
MATRIX
MatrixInvertViaLinearEqSolve(MATRIX A, MATRIX B)
{
    MATRIX result;

    int N; 	// The number of linear equations
    int nrhs;
    int lda;
    int* ipiv;
    int ldb;
    int info;

    N = A.nRow;

    nrhs = B.nCol;

    lda = A.nRow;
    ldb = B.nRow;

	result = newMatrix(B.nRow, B.nCol);

    ipiv = (int *) td_malloc(N * sizeof(int));
	if (ipiv == NULL)
		reportError(OutOfMemory, "MatrixInvertViaLinearEqSolve.ipiv", FATAL);

	// A gets corrupted after linearEqSolve, so make a copy
	double* ADataCopy;
	ADataCopy = (double *) td_malloc(A.nRow * A.nCol * sizeof(double));
	if (ADataCopy == NULL)
		reportError(OutOfMemory, "MatrixInvertViaLinearEqSolve.ADataCopy", FATAL);
	memcpy(ADataCopy, A.data, (A.nRow * A.nCol * sizeof(double)));

	memcpy(result.data, B.data, (B.nRow * B.nCol * sizeof(double)));

	// switch to Column major order
	double *ADataCopyColM;
	ADataCopyColM = toColumnMajor(ADataCopy, A.nRow, A.nCol);
	double *resultColM;
	resultColM = toColumnMajor(result.data, result.nRow, result.nCol);
	dgesv_(&N, &nrhs, ADataCopyColM, &lda, ipiv, resultColM, &ldb, &info);
	td_free(result.data, result.nRow * result.nCol * sizeof(double));
	result.data = toRowMajor(resultColM, result.nRow, result.nCol);

	td_free(resultColM, result.nRow * result.nCol * sizeof(double));
	td_free(ADataCopyColM, A.nRow * A.nCol * sizeof(double));

    td_free(ADataCopy, A.nRow * A.nCol * sizeof(double));
    td_free(ipiv, N * sizeof(int));

    if(info == 0) { /* succeed */
    	;
    }
    else {
    	char msg[MAX_ERROR_MSG_LENGTH];
    	if (info < 0) {
    		// if INFO = -i, the i-th argument had an illegal value
    		snprintf(msg, MAX_ERROR_MSG_LENGTH, "Argument %d had an illegal value", -info);
    	} else {
    		//from description of dgesv.f
    		//  if INFO = i, U(i,i) is exactly zero.  The factorization
    		//                has been completed, but the factor U is exactly
    		//                singular, so the solution could not be computed.
    		char msg1[MAX_ERROR_MSG_LENGTH] = "U(i,i) is exactly zero.  The factorization has been completed, but the factor U is exactly singular, so the solution could not be computed.";
    		snprintf(msg, MAX_ERROR_MSG_LENGTH, "For i = %d, %s", info, msg1);
    	}
    	reportError(DGESVFailed, msg, WARNING);
    	// recover by returning a zero vector
    	result.data = NULL;
    }
	return result;
}

/** Using this function requires first calling linearEqSolve to
 * calculate the LUDecomp of Matrix A.  Solves set of linear equations
 * with multiple right-hand sides, i.e. the number of columns of matrix B.
 * @param[in] A Matrix
 * @param[in] B Matrix
 * @return A Matrix containing the solution x.
 * @ingroup Matrix
 */
MATRIX
LinearEqSolveNRHS(MATRIX A, MATRIX B)
{
    MATRIX result;

    int N; 	// The number of linear equations
    int nrhs;
    int lda;
    int ldb;
    int info;
	char T = 'T';
	char NT = 'N'; // no transpose
	MATRIX resultColM;

    // use LUDecomp for matrix A
    if (LUDecomp.data == NULL) {
    	 reportError(LinearEqSolveFirst, "linearEqSolve.LinearEqSolveNRHS", ERROR);
    }

    N = A.nRow;
    nrhs = B.nCol;
    lda = A.nRow;
    ldb = B.nRow;

		if (getProblemDimension() == 1) {
			resultColM = toColumnMajor2(B);

			// take advantage of intermediate value LUDecomp
			dgetrs_(&T, &N, &nrhs, LUDecomp.data, &lda, IPiv, resultColM.data, &ldb,
					&info);
		} else {
			// 3d problem; complex data
			resultColM = toColumnMajor2(B);

			// remember, no need to transpose LUDecomp matrix, it is already in column major order
			zgetrs_(&NT, &N, &nrhs, LUDecomp.data, &lda, IPiv, resultColM.data, &ldb,
					&info);
		}
	result = toRowMajor2(resultColM);
	// free some memory
	freeMatrix(resultColM);

    if(info == 0) { /* succeed */
    	;
    }
    else {
    	char msg[MAX_ERROR_MSG_LENGTH];
    	if (info < 0) {
    		// if INFO = -i, the i-th argument had an illegal value
    		snprintf(msg, MAX_ERROR_MSG_LENGTH, "Argument %d had an illegal value", -info);
    	} else {
    		snprintf(msg, MAX_ERROR_MSG_LENGTH, "Unknown error %d", info);
    	}
    	reportError(DGETRSFailed, msg, WARNING);
    	// this seems a bit strange to free 'result', then set its values and return it, but
    	// really, we are only freeing the memory pointed to by result.data, nothing else.
    	freeMatrix(result);
    	result.data = NULL;
    }
	return result;
}

/**
 * Free memory allocated for LUDecomp and IPiv intermediate values.
 * @ingroup Matrix
 */
void linearEqSolveCleanup() {
    if (LUDecomp.data != NULL) {
    	freeMatrix(LUDecomp);
    }

    if (IPiv != NULL) {
    	td_free(IPiv, LUDecomp.nRow * sizeof(int));
    }
}
