/*
 * matrix.c
 *
 *  Created on: Apr 22, 2010
 *      Author: Brenda Javornik (e-mail: brendaj (at) cora.nwra.com)
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <complex.h>
#include <time.h>
#include "clapack.h"
#include "utilities.h"
#include "matrix.h"
#include "MatrixFunctionErrorMessages.h"
#include "errorHandler.h"

#define DEBUG 0

#ifndef NAN
#define NAN (0.0/0.0)
#endif

/**
 * @defgroup Matrix Matrix - Matrix creates and manipulates data in 1D and 2D matrix form.
 * @ingroup SupportingModules
 */

/** return the lesser of two values */
#define MIN(X,Y) (( X > Y ) ? Y : X)

/** External LAPACK function */
void dgetrf_(const int *M, const int *N, double *A, const int *lda, int *ipiv, int *info);
/** External LAPACK function */
void zgetrf_(const int *M, const int *N, double complex *A, const int *lda, int *ipiv, int *info);

/** External LAPACK function */
int dgetri_(int *n, double *a, int *lda, int *ipiv, double *work, int *lwork, int *info);
/** External LAPACK function */
int zgetri_(int *n, double complex *a, int *lda, int *ipiv, double complex *work, int *lwork, int *info);

/** External LAPACK function */
int dgemm_( const char *transA, const char *transB, const int *M, const int *N, const int *K, const double *alpha,
		const double *A, const int *ldA, const double *B, const int *ldB, const double *beta, double *C,
		const int *ldC);

/** External LAPACK function */
int zgemm_( const char *transA, const char *transB, const int *M, const int *N, const int *K, const double complex *alpha,
		const double complex *A, const int *ldA, const double complex *B, const int *ldB, const double complex *beta,
		double complex *C, const int *ldC);

/** External LAPACK function */
void dpotrf_(const char *uplo, const int *N, double *A, const int *lda, int *info);
/** External LAPACK function */
void zpotrf_(const char *uplo, const int *N, double complex *A, const int *lda, int *info);

/** External ATLAS BLAS function */
extern void ATL_dgeadd(const int M, const int N, const double alpha,
                 const double *A, const int lda, const double beta,
                 double *C, const int ldc);

extern void ATL_zgeadd(const int M, const int N, const double complex *alpha,
                  const double complex *A, const int lda, const double complex *beta,
                  double complex *C, const int ldc);

VECTOR linearEqSolve(MATRIX A, VECTOR v);

/** Creates a new matrix and allocates space for the data of type double
 * @param[in] r number of rows
 * @param[in] c number of columns
 * @return A new matrix
 * @ingroup Matrix
 */
MATRIX newMatrix(int r, int c) {
	MATRIX A;

	A.data = (double *) td_malloc(r * c * sizeof(double));
	if (A.data == NULL)
		reportError(OutOfMemory, "newMatrix", FATAL);
	A.nRow = r;
	A.nCol = c;
	A.mtype = REAL;
	return A;
}

/** Creates a new matrix and initializes the data to array sent
 * @param[in] r number of rows
 * @param[in] c number of columns
 * @param[in] Idata point to initial data values
 * @return A new matrix
 * @ingroup Matrix
 */
MATRIX newMatrixI(int r, int c, double *Idata) {
	MATRIX A;

	if (Idata == NULL) {
		A.data = (double *) td_malloc(r * c * sizeof(double));
		if (A.data == NULL)
			reportError(OutOfMemory, "newMatrix", FATAL);
	} else {
		A.data = Idata;
	}
	A.nRow = r;
	A.nCol = c;
	A.mtype = REAL;
	return A;
}

/**
 Creates a new matrix of double complex values and allocates space for
 data.
 @param[in]  r  The number of rows
 @param[in]  c  The number of columns
 @return A new matrix
 @ingroup Matrix
 */
MATRIX newComplexMatrix(int r, int c) {
	MATRIX A;

	A.data = (double complex *) td_malloc(r * c * sizeof(double complex));
	if (A.data == NULL)
		reportError(OutOfMemory, "newComplexMatrix", FATAL);
	A.nRow = r;
	A.nCol = c;
	A.mtype = COMPLEX;
	return A;
}

/** Creates a new VECTOR.
 * @param[in] n the size of the vector
 * @return A new VECTOR object.
 * @ingroup Matrix
 */
VECTOR newVector(int n) {
	VECTOR v;

	v.data = (double *) td_malloc(n * sizeof(double));
	if (v.data == NULL)
		reportError(OutOfMemory, "newVector", FATAL);
	v.n = n;
	return v;
}

/**
 * Free memory that stores matrix data. The matrix structure is not released.
 * @param[in] m matrix with data to free.
 * @ingroup Matrix
 */
void freeMatrix(MATRIX m) {
	int factor = sizeof(double);
	if (m.mtype == COMPLEX) {
		factor = sizeof(double complex);
	}
	td_free(m.data, m.nRow * m.nCol * factor);
}

/**
 * Free memory that stores vector data. The vector structure is not released.
 * @param[in] v Vector with data to free.
 * @ingroup Matrix
 */
void freeVector(VECTOR v) {
	int factor = sizeof(double);
	td_free(v.data, v.n * factor);
}

/**
 *  Not used
 */
unsigned int setRandomSeed(unsigned int seed) {

	printf("random seed %u\n", seed);
	srand48(seed);

	return seed;
}

/**
 * Generate Gaussian random numbers; mean of zero, variance of 1.
 * Use a method discussed in Knuth and due originally to Marsaglia:
 * @return Gaussian random number
 */
double gaussrand()
{
	static double V1, V2, S;
	static int phase = 0;
	double X;

	if(phase == 0) {
		do {
			double U1 = drand48();
			double U2 = drand48();

			V1 = 2 * U1 - 1;
			V2 = 2 * U2 - 1;
			S = V1 * V1 + V2 * V2;
			} while(S >= 1 || S == 0);

		X = V1 * sqrt(-2 * log(S) / S);
	} else {
		X = V2 * sqrt(-2 * log(S) / S);
	}
	phase = 1 - phase;

	return X;
}


/**
 * Set seed for random number generator to system time
 * @return Seed value.
 */
unsigned int setRandomSeedToTime() {

	time_t seconds;
	if (time(&seconds) <= -1)
		reportError("System time not available", "newRandomMatrix", FATAL);

	unsigned int seed;
	seed = (unsigned int) seconds;
	printf("random seed %u\n", seed);
	srand48(seed);

	return seed;
}

/** Creates a new matrix of random numbers and allocates space for the data of type
 * double complex.  NOTE: the seed for the random number generator is set separately.
 * @param[in] r number of rows
 * @param[in] c number of columns
 * @return A new matrix
 * @ingroup Matrix
 */
MATRIX newRandomMatrix(int r, int c) {
	MATRIX A;
	A = newComplexMatrix(r, c);

	// fill with random numbers
	int i;
	double complex *p;
	p = A.data;
	for (i=0; i<r*c; i++) {
		*p++ = gaussrand() + 0.0*I;
	}

	return A;
}

/**
 * Multiply two matrices and return a new matrix containing the product of AB.
 * data.
 * @param[in] TransA Letter 'N', 'T', or 'C', for no transpose, transpose, and
 * conjugate transpose.
 * @param[in] TransB Letter 'N', 'T', or 'C', for no transpose, transpose, and
 * conjugate transpose.
 * @param[in]  A  Matrix A
 * @param[in]  B  Matrix B
 * @return A new matrix AB
 * @ingroup Matrix
 */
MATRIX multiplyMM2(char TransA, char TransB, MATRIX A, MATRIX B) {

	double complex alphaC;
	double complex betaC;
	int lda; // first dimension of A
	int ldb; // first dimension of B
	int ldc; // first dimension of C
	int l, m, n;
	char letterC = 'C';
	char letterT = 'T';
	char letterN = 'N';
	char opA, opB;
	double alpha;
	double beta;
	MATRIX C;
	int CRows, CCols;
	BOOLEAN isComplex;

	alpha = 1.0;
	alphaC = 1.0;
	beta = 0.0;
	betaC = beta + 0.0*I;

	if (A.mtype != B.mtype)
		reportError("cannot multiply Real and Complex matrix types", "multiplyMM2", FATAL);

	// convert from row-major order to column-major order
	MATRIX ACol = toColumnMajor2(A);
	MATRIX BCol = toColumnMajor2(B);

	opA = TransA;
	opB = TransB;

	// determine value for m
	if (TransA == letterN) {
		m = A.nCol;
		CRows = A.nRow;
	}
	else {
		m = A.nRow;
		CRows = A.nCol;
	}
	if (TransB == letterN) {
		CCols = B.nCol;
		if (m != B.nRow) reportError("Incompatible matrix sizes","multiplyMM2", FATAL);
	} else {
		CCols = B.nRow;
		if (m != B.nCol) reportError("Incompatible matrix sizes 2","multiplyMM2", FATAL);
	}

	if ((A.mtype == COMPLEX) && (B.mtype == COMPLEX)) {
		isComplex = TRUE;
		C = newComplexMatrix(CRows, CCols);
		if (TransB == letterT)
				opB = letterC;
		if (TransA == letterT)
				opA = letterC;
	} else {
		isComplex = FALSE;
		C = newMatrix(CRows, CCols);
	}

	lda = A.nRow;
	ldb = B.nRow;
	ldc = C.nRow;

	l = C.nRow;
	n = C.nCol;

	if (isComplex) {
		zgemm_(&opA, &opB, &l, &n, &m, &alphaC, ACol.data, &lda, BCol.data, &ldb, &betaC, C.data, &ldc);
	} else {
		dgemm_(&opA, &opB, &l, &n, &m, &alpha, ACol.data, &lda, BCol.data, &ldb, &beta, C.data, &ldc);
	}

	freeMatrix(ACol);
	freeMatrix(BCol);

	// convert from column-major order to row-major order
	MATRIX result = toRowMajor2(C);
	freeMatrix(C);

	return result;
}

/* HACK!  I need a form of multiplyMM2 without a forced conjugate on a transpose
 * when the matrix is complex
 *
 */
/**
 * Multiply two matrices and return a new matrix containing the product of AB.
 * data.
 * @param[in] TransA Letter 'N', 'T', or 'C', for no transpose, transpose, and
 * conjugate transpose.
 * @param[in] TransB Letter 'N', 'T', or 'C', for no transpose, transpose, and
 * conjugate transpose.
 * @param[in]  A  Matrix A
 * @param[in]  B  Matrix B
 * @return A new matrix AB
 * @ingroup Matrix
 */
MATRIX multiplyMM2NoForceC(char TransA, char TransB, MATRIX A, MATRIX B) {

	double complex alphaC;
	double complex betaC;
	int lda; // first dimension of A
	int ldb; // first dimension of B
	int ldc; // first dimension of C
	int l, m, n;
	char letterN = 'N';
	char opA, opB;
	double alpha;
	double beta;
	MATRIX C;
	int CRows, CCols;
	BOOLEAN isComplex;

	alpha = 1.0;
	alphaC = 1.0;
	beta = 0.0;
	betaC = beta + 0.0*I;

	if (A.mtype != B.mtype)
		reportError("cannot multiply Real and Complex matrix types", "multiplyMM2", FATAL);

	// convert from row-major order to column-major order
	MATRIX ACol = toColumnMajor2(A);
	MATRIX BCol = toColumnMajor2(B);

	opA = TransA;
	opB = TransB;

	// determine value for m
	if (TransA == letterN) {
		m = A.nCol;
		CRows = A.nRow;
	}
	else {
		m = A.nRow;
		CRows = A.nCol;
	}
	if (TransB == letterN) {
		CCols = B.nCol;
		if (m != B.nRow) reportError("Incompatible matrix sizes","multiplyMM2", FATAL);
	} else {
		CCols = B.nRow;
		if (m != B.nCol) reportError("Incompatible matrix sizes 2","multiplyMM2", FATAL);
	}

	if ((A.mtype == COMPLEX) && (B.mtype == COMPLEX)) {
		isComplex = TRUE;
		C = newComplexMatrix(CRows, CCols);
		/* don't force conjugate with a transpose
		if (TransB == letterT)
				opB = letterC;
		if (TransA == letterT)
				opA = letterC;
		*/
	} else {
		isComplex = FALSE;
		C = newMatrix(CRows, CCols);
	}

	lda = A.nRow;
	ldb = B.nRow;
	ldc = C.nRow;

	l = C.nRow;
	n = C.nCol;

	if (isComplex) {
		zgemm_(&opA, &opB, &l, &n, &m, &alphaC, ACol.data, &lda, BCol.data, &ldb, &betaC, C.data, &ldc);
	} else {
		dgemm_(&opA, &opB, &l, &n, &m, &alpha, ACol.data, &lda, BCol.data, &ldb, &beta, C.data, &ldc);
	}

	freeMatrix(ACol);
	freeMatrix(BCol);

	// convert from column-major order to row-major order
	MATRIX result = toRowMajor2(C);
	freeMatrix(C);

	return result;
}

// KTheta = multiplyMMFull('C', 'N', (2.0*M_PI)*(2.0*M_PI), 1.0, ttd, Theta);
/**
 * Multiply two matrices and return a new matrix containing the product of AB.
 * data.  C = alpha A x B + beta C
 * @param[in] TransA Letter 'N', 'T', or 'C', for no transpose, transpose, and
 * conjugate transpose.
 * @param[in] TransB Letter 'N', 'T', or 'C', for no transpose, transpose, and
 * conjugate transpose.
 * @param[in] alpha Multiply matrix A by scalar alpha.
 * @param[in] beta Multiply matrix C by scalar beta.
 * @param[in]  A  Matrix A
 * @param[in]  B  Matrix B
 * @return A new matrix AB
 * @ingroup Matrix
 */
MATRIX multiplyMMFull(char TransA, char TransB, double alpha, double beta, MATRIX A, MATRIX B) {

	double complex alphaC;
	double complex betaC;
	int lda; // first dimension of A
	int ldb; // first dimension of B
	int ldc; // first dimension of C
	int l, m, n;
	char letterC = 'C';
	char letterT = 'T';
	char letterN = 'N';
	char opA, opB;
	MATRIX C;
	int CRows, CCols;
	BOOLEAN isComplex;

	alphaC = alpha + 0.0*I;
	betaC = beta + 0.0*I;

	if (A.mtype != B.mtype)
		reportError("cannot multiply Real and Complex matrix types", "multiplyMMFull", FATAL);

	// convert from row-major order to column-major order
	MATRIX ACol = toColumnMajor2(A);
	MATRIX BCol = toColumnMajor2(B);

	opA = TransA;
	opB = TransB;

	// determine value for m
	if (TransA == letterN) {
		m = A.nCol;
		CRows = A.nRow;
	}
	else {
		m = A.nRow;
		CRows = A.nCol;
	}
	if (TransB == letterN) {
		CCols = B.nCol;
		if (m != B.nRow) reportError("Incompatible matrix sizes","multiplyMM2", FATAL);
	} else {
		CCols = B.nRow;
		if (m != B.nCol) reportError("Incompatible matrix sizes 2","multiplyMM2", FATAL);
	}

	if ((A.mtype == COMPLEX) && (B.mtype == COMPLEX)) {
		isComplex = TRUE;
		C = newComplexMatrix(CRows, CCols);
		if (TransB == letterT)
				opB = letterC;
		if (TransA == letterT)
				opA = letterC;
	} else {
		isComplex = FALSE;
		C = newMatrix(CRows, CCols);
	}

	lda = A.nRow;
	ldb = B.nRow;
	ldc = C.nRow;

	l = C.nRow;
	n = C.nCol;

	if (isComplex) {
		zgemm_(&opA, &opB, &l, &n, &m, &alphaC, ACol.data, &lda, BCol.data, &ldb, &betaC, C.data, &ldc);
	} else {
		dgemm_(&opA, &opB, &l, &n, &m, &alpha, ACol.data, &lda, BCol.data, &ldb, &beta, C.data, &ldc);
	}

	freeMatrix(ACol);
	freeMatrix(BCol);

	// convert from column-major order to row-major order
	MATRIX result = toRowMajor2(C);
	freeMatrix(C);

	return result;
}

/** multiply matrix by a scalar value
 * @param[in] m A Matrix.
 * @param[in] scalar A single value.
 * @return The resulting matrix.
 * @ingroup Matrix
 */
MATRIX multiplyScalar(MATRIX m, double scalar) {

	int r;
	MATRIX result;
	if (m.mtype == COMPLEX) {
		double complex *p, *q;
		result = newComplexMatrix(m.nRow, m.nCol);
		p = result.data;
		q = m.data;
		for (r = 0; r < m.nRow * m.nCol; r++) {
			*p++ = *q++ * scalar;
		}
	} else {
		double *p, *q;
		result = newMatrix(m.nRow, m.nCol);
		p = result.data;
		q = m.data;
		for (r = 0; r < m.nRow * m.nCol; r++) {
			*p++ = *q++ * scalar;
		}
	}
	return result;

}

/** multiply matrix by a scalar value
 * @param[in] m A Matrix.
 * @param[in] scalar A single value.
 * @return The resulting matrix.
 * @ingroup Matrix
 */
MATRIX multiplyScalarZ(MATRIX m, double complex scalar) {

	int r;
	MATRIX result;
	double complex *p;
	result = newComplexMatrix(m.nRow, m.nCol);
	p = result.data;

	if (m.mtype == COMPLEX) {
		double complex *q;
		q = m.data;
		for (r = 0; r < m.nRow * m.nCol; r++) {
			*p++ = *q++ * scalar;
		}
	} else {
		double *qf;
		qf = m.data;
		//double *qq;
		//qq = m.data;
		//printf("sizeof(double) = = %d", sizeof(*qf));
		for (r = 0; r < m.nRow * m.nCol; r++) {
			//printf("p=%lf + %lfi, %1d: q=%lf vs arrayidx = %lf\n", creal(*p), cimag(*p), r, *qf, qq[r]);
			//*p++ = *qf++ * scalar;  // original  compiler bug;  qf is being incremented by 16 instead of 8
			// begin patch for compiler bug
			*p++ = *qf * scalar;
			qf++;
			// end patch
			// *p++ = q[r] * scalar; // works
			// printf("*p = %lf + %lfi\n", creal(*(p-1)), cimag(*(p-1)));
		}

	}
	return result;
}
// Element by element multiplication. If both operands are matrices, the number of rows and columns must both agree.
MATRIX starMultiply(MATRIX P, MATRIX Q) {
	MATRIX PQ;

	if ((P.nCol != Q.nCol) || (P.nRow != Q.nRow))
		reportError(MatrixStarMultiplySameSize, "starMultiply", FATAL);
	if ((P.mtype == Q.mtype) && (P.mtype == COMPLEX)) {

		PQ = newComplexMatrix(P.nRow, P.nCol);
		double complex *pp, *pq, *ppq;
		pp = P.data;
		pq = Q.data;
		ppq = PQ.data;
		int i;
		for (i = 0; i < P.nRow * P.nCol; i++)
			*ppq++ = *pp++ * *pq++;
	} else 	if ((P.mtype == Q.mtype) && (P.mtype == REAL)) {
		PQ = newMatrix(P.nRow, P.nCol);
		double  *pp, *pq, *ppq;
		pp = P.data;
		pq = Q.data;
		ppq = PQ.data;
		int i;
		for (i = 0; i < P.nRow * P.nCol; i++)
			*ppq++ = *pp++ * *pq++;
	} else
		reportError(MatrixStarMultiplySameType, "starMultiply", FATAL);

	return PQ;
}

/**
 * Copy a 2D array in column-major layout return the copy in row-major layout.
 * @param[in] A The array of data in column-major layout.
 * @param[in] R The number of rows.
 * @param[in] C The number of columns.
 * @return A 2D array in row-major layout.
 * @ingroup Matrix
 */
// Pass r & c separately instead of inside a MATRIX object because,
// sometimes from a LaPack call all we have is an array and its dimensions.
double* toRowMajor(double* A, int R, int C) {
	// convert from column-major to row-major order
	double* AT;
	int j;

	AT = (double *) td_malloc(R * C * sizeof(double));
	for (j = 0; j < R * C; j++) {
		AT[(j % R) * C + j / R] = A[j];
	}
	return AT;
}

/**
 * Copy a 2D array in row-major layout return the copy in column-major layout.
 * @param[in] AT The array of data in row-major layout.
 * @param[in] R The number of rows.
 * @param[in] C The number of columns.
 * @return A 2D array in column-major layout.
 * @ingroup Matrix
 */
// Returns a copy of the matrix stored in column major order
double* toColumnMajor(double* AT, int R, int C) {
	// convert from row-major to column-major order
	double* ATA;
	int j;
	ATA = (double *) td_malloc(R * C * sizeof(double));
	for (j = 0; j < R * C; j++) {
		ATA[(j % C) * R + j / C] = AT[j];
	}
	return ATA;
}

/**
 Copy a matrix stored in row-major order and convert it to column-major order.
 @param[in]  m  A row-major matrix
 @return A copy of matrix m with column-major layout of data.
 @ingroup Matrix
 */
MATRIX toColumnMajor2(MATRIX m) {
	MATRIX colm;
	int j;
	int R,C;

	R = m.nRow;
	C = m.nCol;

	if (m.mtype == REAL) {
		colm = newMatrix(m.nRow, m.nCol);
		double *p, *q;
		p = m.data;
		q = colm.data;
		for (j = 0; j < m.nRow * m.nCol; j++) {
			q[(j % C) * R + j / C] = p[j];
		}
	} else {
		colm = newComplexMatrix(m.nRow, m.nCol);
		double complex *p, *q;
		p = m.data;
		q = colm.data;
		for (j = 0; j < m.nRow * m.nCol; j++) {
			q[(j % C) * R + j / C] = p[j];
		}
	}
	return colm;
}

/**
 Copy a matrix stored in column-major order and convert it to row-major order.
 @param[in]  m  A column-major matrix
 @return A copy of matrix n with row-major layout of data.
 @ingroup Matrix
 */
MATRIX toRowMajor2(MATRIX m) {
	MATRIX rowm;
	int j;
	int R,C;

	R = m.nRow;
	C = m.nCol;
	if (m.mtype == REAL) {
		rowm = newMatrix(m.nRow, m.nCol);
		double *p, *q;
		p = m.data;
		q = rowm.data;
		for (j = 0; j < m.nRow * m.nCol; j++) {
			q[(j % R) * C + j / R] = p[j];
		}
	} else {
		rowm = newComplexMatrix(m.nRow, m.nCol);
		double complex *p, *q;
		p = m.data;
		q = rowm.data;
		for (j = 0; j < m.nRow * m.nCol; j++) {
			q[(j % R) * C + j / R] = p[j];
		}
	}
	return rowm;
}

MATRIX multiplyMMplusCZ(MATRIX A, MATRIX B, double beta, MATRIX C) {
	// call the appropriate lapack function
	// will need to convert from row-major order to column-major order
	// ZGEMM(TRANSA,TRANSB,M,N,K,ALPHA,A,LDA,B,LDB,BETA,C,LDC)
	/*
	 *  ZGEMM  performs one of the matrix-matrix operations
	 *
	 *     C := alpha*op( A )*op( B ) + beta*C,
	 *
	 *  where  op( X ) is one of
	 *
	 *     op( X ) = X   or   op( X ) = X',
	 *
	 *  alpha and beta are scalars, and A, B and C are matrices, with op( A )
	 *  an m by k matrix,  op( B )  a  k by n matrix and  C an m by n matrix.
	 *
	 */

	double complex alpha;
	double complex betaC;
	int lda; // first dimension of A
	int ldb; // first dimension of B
	int ldc; // first dimension of C
	char letterT = 'T';
	char letterN = 'N';

	if ((A.nCol != B.nCol) || (C.nRow != A.nRow) || (C.nCol != B.nRow))
		reportError("incompatible matrix dimensions", "multiplyMMplusCZ", FATAL);

	alpha = 1.0;
	betaC = beta + 0.0*I;
	lda = A.nRow;
	ldb = B.nRow;
	ldc = C.nRow;

	// convert from row-major order to column-major order
	MATRIX ACol = toColumnMajor2(A);
	MATRIX BCol = toColumnMajor2(B);
	MATRIX CCol = toColumnMajor2(C);

	/*
	 * transa = 'N'
	 * transb = 'T'
	 * l = # of rows in C
	 * n = number of columns in C
	 * m = If transa = 'N', it is the number of columns in matrix A.
	 *     If transb = 'T' or 'C', it is the number of columns in matrix B.
	 * alpha = scalar
	 * a = Matrix A. If transa = 'N', A is used in the computation, and A has l rows and m columns.
	 * lda = If transa = 'N', lda ≥ l.
	 * b = If transb = 'T', BT is used in the computation, and B has n rows and m columns.
	 *     If transb = 'T' or 'C', its size must be ldb by (at least) m.
	 * ldb = If transb = 'T' or 'C', ldb ≥ n.
	 * beta
	 * c = is the l by n matrix C.
	 * ldc = is the leading dimension of the array specified for c. Specified as: an integer; ldc > 0 and ldc ≥ l.
	 */
	zgemm_(&letterN, &letterT,
			&C.nRow, &C.nCol, &A.nCol, &alpha, ACol.data, &lda,
			BCol.data, &ldb, &betaC, CCol.data, &ldc);

	// convert from column-major order to row-major order
	freeMatrix(ACol);
	freeMatrix(BCol);

	MATRIX result = toRowMajor2(CCol);
	freeMatrix(CCol);

	return result;
}

/**
 * Attempt to generalize matrix multiplication and make this really
 * a facade for the lapack library dgemm and zgemm
 */
MATRIX multiplyAlphaABplusC(char TransA, char TransB, MATRIX A, MATRIX B, double alpha, double beta, MATRIX C) {
     //MATRIX multiplyABplusC(char TransA, char TransB, MATRIX A, MATRIX B, double beta, MATRIX C) {
	// call the appropriate lapack function
	// will need to convert from row-major order to column-major order
	// ZGEMM(TRANSA,TRANSB,M,N,K,ALPHA,A,LDA,B,LDB,BETA,C,LDC)
	/*
	 *  ZGEMM  performs one of the matrix-matrix operations
	 *
	 *     C := alpha*op( A )*op( B ) + beta*C,
	 *
	 *  where  op( X ) is one of
	 *
	 *     op( X ) = X   or   op( X ) = X',
	 *
	 *  alpha and beta are scalars, and A, B and C are matrices, with op( A )
	 *  an m by k matrix,  op( B )  a  k by n matrix and  C an m by n matrix.
	 *
	 */

	double complex alphaC;
	double complex betaC;
	int lda; // first dimension of A
	int ldb; // first dimension of B
	int ldc; // first dimension of C
	char letterN = 'N';
	char opA, opB;
	double betaTemp;

	alphaC = alpha + 0.0*I;
	betaC = beta + 0.0*I;
	lda = A.nRow;
	ldb = B.nRow;
	ldc = C.nRow;

	// convert from row-major order to column-major order
	MATRIX ACol = toColumnMajor2(A);
	MATRIX BCol = toColumnMajor2(B);
	MATRIX CCol = toColumnMajor2(C);

	int l = C.nRow;
	int n = C.nCol;
	int m;

	opA = TransA;
	opB = TransB;

	// determine value for m
	if (TransA == letterN) m = A.nCol;
	else m = A.nRow;
	if (TransB == letterN) {
		if (m != B.nRow) reportError("error 1","multiplyABplusC", FATAL);
	} else {
		if (m != B.nCol) reportError("error 2","multiplyABplusC", FATAL);
	}

	if ((A.mtype == COMPLEX) && (B.mtype == COMPLEX)) {
		zgemm_(&opA, &opB, &l, &n, &m, &alphaC, ACol.data, &lda, BCol.data, &ldb, &betaC, CCol.data, &ldc);
	} else {
		betaTemp = beta;
		dgemm_(&opA, &opB, &l, &n, &m, &alpha, ACol.data, &lda, BCol.data, &ldb, &betaTemp, CCol.data, &ldc);
	}

	freeMatrix(ACol);
	freeMatrix(BCol);

	// convert from column-major order to row-major order
	MATRIX result = toRowMajor2(CCol);
	freeMatrix(CCol);

	return result;
}


		/** ORIGINAL **/
/**
 * Attempt to generalize matrix multiplication and make this really
 * a facade for the lapack library dgemm and zgemm
 */

MATRIX multiplyABplusC(char TransA, char TransB, MATRIX A, MATRIX B, double beta, MATRIX C) {

	// call the appropriate lapack function
	// will need to convert from row-major order to column-major order
	// ZGEMM(TRANSA,TRANSB,M,N,K,ALPHA,A,LDA,B,LDB,BETA,C,LDC)
	/*
	 *  ZGEMM  performs one of the matrix-matrix operations
	 *
	 *     C := alpha*op( A )*op( B ) + beta*C,
	 *
	 *  where  op( X ) is one of
	 *
	 *     op( X ) = X   or   op( X ) = X',
	 *
	 *  alpha and beta are scalars, and A, B and C are matrices, with op( A )
	 *  an m by k matrix,  op( B )  a  k by n matrix and  C an m by n matrix.
	 *
	 */

	double complex alphaC;
	double complex betaC;
	int lda; // first dimension of A
	int ldb; // first dimension of B
	int ldc; // first dimension of C
	char letterC = 'C';
	char letterT = 'T';
	char letterN = 'N';
	char opA, opB;
	double alpha;
	double betaTemp;

	alpha = 1.0;
	alphaC = 1.0;
	betaC = beta + 0.0*I;
	lda = A.nRow;
	ldb = B.nRow;
	ldc = C.nRow;

	// convert from row-major order to column-major order
	MATRIX ACol = toColumnMajor2(A);
	MATRIX BCol = toColumnMajor2(B);
	MATRIX CCol = toColumnMajor2(C);

	int l = C.nRow;
	int n = C.nCol;
	int m;

	opA = TransA;
	opB = TransB;

	// determine value for m
	if (TransA == letterN) m = A.nCol;
	else m = A.nRow;
	if (TransB == letterN) {
		if (m != B.nRow) reportError("error 1","multiplyABplusC", FATAL);
	} else {
		if (m != B.nCol) reportError("error 2","multiplyABplusC", FATAL);
	}

	if ((TransB == letterT) && (A.mtype == COMPLEX) && (B.mtype == COMPLEX))
			opB = letterC;
	if ((TransA == letterT) && (A.mtype == COMPLEX) && (B.mtype == COMPLEX))
			opA = letterC;


	if ((A.mtype == COMPLEX) && (B.mtype == COMPLEX)) {
		zgemm_(&opA, &opB, &l, &n, &m, &alphaC, ACol.data, &lda, BCol.data, &ldb, &betaC, CCol.data, &ldc);
	} else {
		betaTemp = beta;
		dgemm_(&opA, &opB, &l, &n, &m, &alpha, ACol.data, &lda, BCol.data, &ldb, &betaTemp, CCol.data, &ldc);
	}

	freeMatrix(ACol);
	freeMatrix(BCol);

	// convert from column-major order to row-major order
	MATRIX result = toRowMajor2(CCol);
	freeMatrix(CCol);

	return result;
}


   /** end ORIGINAL **/

/**
 * Attempt to generalize matrix multiplication and make this really
 * a facade for the lapack library dgemm and zgemm
 */
/*
MATRIX multiplyABplusC(char TransA, char TransB, MATRIX A, MATRIX B, double beta, MATRIX C) {
	double alpha = 1.0;
	return multiplyAlphaABplusC( TransA,  TransB,  A,  B,  alpha,  beta,  C);
}
*/

/**
 Convert a fits2d object to a Matrix object.
 @param[in]  f  A fits2d object; i.e. a 2-d array with different member names
 and always data of type double.
 @return The converted object as a matrix.
 @ingroup Matrix
 */
MATRIX toMatrix(fits2d f) {
	MATRIX m;
	m.data = f.data;
	m.nCol = f.yDim;
	m.nRow = f.xDim;
	m.mtype = REAL;
	return m;
}

/**
 Convert a fits2d object to a Vector object.
 @param[in]  f  A fits2d object; i.e. a 2-d array with different member names
 and always data of type double.
 @return The converted object as a vector.
 @ingroup Matrix
 */
VECTOR toVector(fits2d f) {
	VECTOR v;
	v.data = f.data;
	if (f.xDim == 1) {
		v.n = f.yDim;
	}
	else if (f.yDim == 1) {
		v.n = f.xDim;
	} else {
		v.data = NULL;
		v.n = -1;
		reportError(ToVectorFailes, "matrix.toVector()", ERROR);
	}

	return v;
}

/**
 * Convert from a Vector object to a fits2d object.
 * @param[in]  v  A vector
 * @return The converted vector as a fits2d object; i.e. a 2-d array with different member names
 * and always data of type double.
 * @ingroup Matrix
 */
fits2d toFitsObject(VECTOR v) {
	fits2d f;
	f.data = v.data;
	f.xDim = v.n;
	f.yDim = 1;
	return f;
}

/** Convert from a Matrix object to a fits2d object.
 * @param[in]  m  A matrix.
 * @return The converted matrix as a fits2d object; i.e. a 2-d array with different member names
 * and always data of type double.
 * @ingroup Matrix
 */
fits2d MatrixToFitsObject(MATRIX m) {
	if (m.mtype == COMPLEX)
		reportError("Cannot convert complex matrix to fits2d", "MatrixToFitsObject", FATAL);
	fits2d f;
	f.data = m.data;
	f.xDim = m.nRow;
	f.yDim = m.nCol;
	return f;
}

// Keep this because it writes a complex matrix to a fits file
/** Convert from a Matrix object to a fits3d object.
 * @param[in] m A matrix.
 * @return The converted matrix as a fits3d object.
 * @ingroup Matrix
 */
fits3d MatrixToFits3dObject(MATRIX m) {
	fits3d f;
	int r,c;
	double complex *p;
	int x;
	int idx;

	// reminder of fits3d object ...
	//typedef struct {
	//  double **Real;
	//  double **Imaginary;
	//  int xDim;
	//  int yDim;
	//  int zDim;
	//} fits3d;


	if (m.mtype != COMPLEX)
		reportError("only complex matrix can be converted to fits3", "", FATAL);
	//f.Real = (double **) td_malloc(m.nRow * m.nCol * sizeof(double*))
	// get memory for each row
	f.Real = (double **) td_malloc(m.nRow * sizeof(double*));
	// get memory for each column
	for (x = 0; x < m.nRow; x++) {
		f.Real[x] = (double *) td_malloc(m.nCol * sizeof(double));
	}

	// get memory for each row
	f.Imaginary = (double **) td_malloc(m.nRow * sizeof(double*));
	// get memory for each column
	for (x = 0; x < m.nRow; x++) {
		f.Imaginary[x] = (double *) td_malloc(m.nCol * sizeof(double));
	}

	p = m.data;
	for (r=0; r<m.nRow; r++) {
		for (c=0; c<m.nCol; c++) {
			idx = r*m.nCol + c;
			f.Real[r][c] = creal(p[idx]);
			f.Imaginary[r][c] = cimag(p[idx]);
		}
	}

	//f.data = m.data;
	f.xDim = m.nRow;
	f.yDim = m.nCol;
	f.zDim = 2;
	return f;
}

/**
 * Expand a compact representation of a diagonal matrix.
 * @param[in] v A vector containing the diagonal elements of a matrix.
 * @return A diagonal matrix with diagonal elements taken from input vector.
 * @ingroup Matrix
 */
MATRIX expandDiagonalMatrix(VECTOR v) {
	MATRIX d;
	int i;
	int spacer;
	double *p;

	d = newMatrix(v.n, v.n);
	spacer = v.n + 1;
	p = d.data;
	for (i = 0; i < v.n*v.n; i++) {
		if ((i % spacer) != 0) {
			//d.data[i] = 0.0;
			p[i] = 0.0;
		} else {
			//d.data[i] = v.data[i/spacer];
			p[i] = v.data[i/spacer];
		}
	}
	return d;
}

/**
 * Transpose a matrix.
 * @param[in] m A matrix.
 * @return The transpose of the matrix.
 * @ingroup Matrix
 */
MATRIX transpose(MATRIX m) {
	MATRIX MT;
	int r,c;

	if (m.mtype == REAL) {
		MT = newMatrix(m.nCol, m.nRow);
		double *p = (double *) m.data;
		double *q = (double *) MT.data;
		for (r=0; r<m.nRow; r++) {
			for (c=0; c<m.nCol; c++ ){
				q[c*MT.nCol + r] = p[r*m.nCol + c];
			}
		}
	} else {
		MT = newComplexMatrix(m.nCol, m.nRow);
		double complex *p = (double complex*) m.data;
		double complex *q = (double complex *) MT.data;
		for (r=0; r<m.nRow; r++) {
			for (c=0; c<m.nCol; c++ ){
				q[c*MT.nCol + r] = p[r*m.nCol + c];
			}
		}
	}

	return MT;
}

/**
 * Take the complex conjugate of a matrix.  Returns a new matrix.
 * @param[in] m Take complex conjugate of this matrix.
 * @return New matrix containing conjugate of original matrix.
 * @ingroup Matrix
 */
MATRIX conjugate(MATRIX m) {
	MATRIX CJ;
	int r;

	if (m.mtype == REAL) {
		reportError(MatrixNoConjugateReal, "conjugate", FATAL);
	} else {
		CJ = newComplexMatrix(m.nRow, m.nCol);
		double complex *p = (double complex*) m.data;
		double complex *q = (double complex *) CJ.data;
		for (r=0; r<m.nRow * m.nCol; r++) {
				q[r] = conj(p[r]);
		}
	}

	return CJ;
}

/**
 * Return a copy of a matrix.  Returns a new matrix.
 * @param[in] m A matrix.
 * @return New matrix containing copy of original matrix.
 * @ingroup Matrix
 */
MATRIX matrix_copy(MATRIX m) {
	MATRIX c;

	if (m.mtype == REAL) {
		c = newMatrix(m.nRow, m.nCol);
		double  *p = (double *) m.data;
		double  *q = (double  *) c.data;
		// memcpy(to, from, size)
		memcpy(q, p, (m.nCol * m.nRow * sizeof(double)));
	} else {
		c = newComplexMatrix(m.nRow, m.nCol);
		double complex *p = (double complex*) m.data;
		double complex *q = (double complex *) c.data;
		// memcpy(to, from, size)
		memcpy(q, p, (m.nCol * m.nRow * sizeof(double complex)));
	}
	return c;
}


/**
 * Inverse of a diagonal matrix is 1/aii, where aii are the diagonal elements
 * @param[in] m A diagonal matrix.
 * @return The inverse of the matrix
 * @ingroup Matrix
 */
MATRIX inverseD(MATRIX m) {
	MATRIX result;
	double *p, *q;

	if (m.mtype != REAL) reportError(RealMatrixExpected, "matrix.inverseD", FATAL);

	if (m.nCol == m.nRow) {
		int i;
		result = newMatrix(m.nRow, m.nCol);
		p = result.data;
		q = m.data;
		// initialize result data to zero
		for (i = 0; i<m.nRow * m.nCol; i++) {
			//result.data[i] = 0.0;
			p[i] = 0.0;
		}
		// now fill in the diagonal elements
		p = result.data;
		q = m.data;
		for (i=0; i<m.nRow; i++) {
			if (*q == 0.0) {
				reportError(DivByZero, " matrix.inverseD()", ERROR);
				*p = NAN;
			}
			else
				*p = 1.0/(*q);
			p += m.nRow + 1;
			q += m.nRow + 1;
		}

	}
	return result;
}

/**
 * Inverse of a general matrix.
 * @param[in] m A  matrix.
 * @return The inverse of the matrix.  If an error occurs during the inversion,
 * the matrix returned has data = NULL, and dimensions are -1.
 * @ingroup Matrix
 */
MATRIX inverse(MATRIX m) {
	MATRIX MI;
	//double *MColMajor;
	MATRIX MColMajor;
	int lda;
	int *ipiv;
	int info;

	//MI = newEmptyMatrix(m.nRow, m.nCol);
	// make a copy of m because the inverse is returned in the passed matrix
	// toColumnMajor returns a copy of the matrix
	MColMajor = toColumnMajor2(m);
	lda = m.nRow;
	ipiv = (int *) td_malloc(m.nCol * m.nRow * sizeof(int));
	// A = m
	// call DLANGE to compute the norm of A(1,infinity)
	// call DGETRF to compute LU factorization
	if (m.mtype == REAL)
		dgetrf_(&(m.nRow), &(m.nCol), MColMajor.data, &lda, ipiv, &info);
	else
		zgetrf_(&(m.nRow), &(m.nCol), MColMajor.data, &lda, ipiv, &info);

	if (info != 0) {
		// problem with LU factorization of matrix
	    char msg[MAX_ERROR_MSG_LENGTH];
		if (info < 0) {
			// if INFO = -i, the i-th argument had an illegal value
			snprintf(msg, MAX_ERROR_MSG_LENGTH, "Argument %d had an illegal value", -info);
		} else {
			//from description of dgetrf.f
			//  if INFO = i, U(i,i) is exactly zero.  The factorization
			//                has been completed, but the factor U is exactly
			//                singular, and division by zero will occur if it is used
			//                to solve a system of equations.

			char msg1[MAX_ERROR_MSG_LENGTH] = "U(i,i) is exactly zero.  The factorization has been completed, but the factor U is exactly singular, and division by zero will occur if it is used to solve a system of equations.";
			snprintf(msg, MAX_ERROR_MSG_LENGTH, "For i = %d, %s", info, msg1);
		}
		reportError(DGETRFFailed, msg, ERROR);

		MI.data = NULL;
		MI.nCol = -1;
		MI.nRow = -1;
	} else { // continue calculation
		if (m.mtype == REAL) {
			// call DGECON to compute 1/cond(A)
			// call DGETRI to compute the inverse of A using the LU factorization
			// 		computed by DGETRF
			int lwork = -1;
			double *work;
			double worksize[1];

			// query for the optimal size of the work array

			dgetri_(&(m.nRow), MColMajor.data, &lda, ipiv, worksize, &lwork, &info);
			// call again sending the optimal size of the work array
			// and actually perform the operation
			work = (double *) td_malloc((int) worksize[0] * sizeof(double));
			lwork = worksize[0];
			dgetri_(&(m.nRow), MColMajor.data, &lda, ipiv, work, &lwork, &info);
			td_free(work, (int) worksize[0] * sizeof(double));
		} else {
			// Complex data
			// call DGECON to compute 1/cond(A)
			// call DGETRI to compute the inverse of A using the LU factorization
			// 		computed by DGETRF
			int lwork = -1;
			double complex *work;
			double complex worksize[1];

			// query for the optimal size of the work array

			zgetri_(&(m.nRow), MColMajor.data, &lda, ipiv, worksize, &lwork, &info);
			// call again sending the optimal size of the work array
			// and actually perform the operation
			work = (double complex *) td_malloc((int) creal(worksize[0]) * sizeof(double complex));
			lwork = worksize[0];
			zgetri_(&(m.nRow), MColMajor.data, &lda, ipiv, work, &lwork, &info);
			td_free(work, lwork * sizeof(double complex));
		}
		if (info != 0) {
			// problem with computing inverse
		    char msg[MAX_ERROR_MSG_LENGTH];
			if (info < 0) {
				// if INFO = -i, the i-th argument had an illegal value
				snprintf(msg, MAX_ERROR_MSG_LENGTH, "Argument %d had an illegal value", -info);
			} else {
				//from description of dgetri.f
				//  if INFO = i, U(i,i) is exactly zero.  the matrix is
				//               singular and its inverse could not be computed.
				char msg1[MAX_ERROR_MSG_LENGTH] = "U(i,i) is exactly zero.  The matrix is singular and its inverse could not be computed.";
				snprintf(msg, MAX_ERROR_MSG_LENGTH, "For i = %d, %s", info, msg1);
			}
			reportError(DGETRIFailed, msg, ERROR);
			MI.data = NULL;
			MI.nCol = -1;
			MI.nRow = -1;
		} else {
			MI = toRowMajor2(MColMajor);
		}
	}

	td_free(ipiv, m.nCol * m.nRow * sizeof(int));
	if (MColMajor.mtype == REAL) {
		td_free(MColMajor.data, MColMajor.nRow * MColMajor.nCol * sizeof(double));
	} else {
		td_free(MColMajor.data, MColMajor.nRow * MColMajor.nCol * sizeof(double complex));
	}
	return MI;
}

/**
 * Cholesky factorization of a symmetric, positive, definite matrix.
 * @param[in] m is a matrix to factor.
 * @return An upper triangular matrix that is the Cholesky factorization of the matrix, m.
 * @ingroup Matrix
 */
MATRIX Cholesky(MATRIX m) {
	MATRIX MI;
	MATRIX MColMajor;
	int lda;
	int info;
	char letterU = 'U';

	// make a copy of m because the inverse is returned in the passed matrix
	// toColumnMajor returns a copy of the matrix
	MColMajor = toColumnMajor2(m);
	lda = m.nRow;

	// call xPOTRF to compute Cholesky factorization
	if (m.mtype == REAL)
		// void dpotrf_(const char *uplo, const int *N, double *A, const int *lda, int *info);
		dpotrf_(&letterU, &(m.nCol), MColMajor.data, &lda, &info);
	else
		zpotrf_(&letterU, &(m.nCol), MColMajor.data, &lda, &info);

	if (info != 0) {
		// problem with LU factorization of matrix
	    char msg[MAX_ERROR_MSG_LENGTH];
		if (info < 0) {
			// if INFO = -i, the i-th argument had an illegal value
			snprintf(msg, MAX_ERROR_MSG_LENGTH, "Argument %d had an illegal value", -info);
		} else {
			//from description of dgetrf.f
			//  if INFO = i, U(i,i) is exactly zero.  The factorization
			//                has been completed, but the factor U is exactly
			//                singular, and division by zero will occur if it is used
			//                to solve a system of equations.

			char msg1[MAX_ERROR_MSG_LENGTH] = "the leading minor of order i is not positive definite, and the factorization could not be completed.";
			snprintf(msg, MAX_ERROR_MSG_LENGTH, "For i = %d, %s", info, msg1);
		}
		reportError(DPOTRFFailed, msg, ERROR);

		MI.data = NULL;
		MI.nCol = -1;
		MI.nRow = -1;
	} else { // continue calculation
		MI = toRowMajor2(MColMajor);
	}

	if (MColMajor.mtype == REAL) {
		td_free(MColMajor.data, MColMajor.nRow * MColMajor.nCol * sizeof(double));
	} else {
		td_free(MColMajor.data, MColMajor.nRow * MColMajor.nCol * sizeof(double complex));
	}
	return MI;
}


/**
 * Convert a complex matrix to a real matrix.  If the imaginary
 * part of the complex matrix elements is nearly zero, the conversion
 * proceeds.  Otherwise, an error is thrown.
 * @param[in] m A complex matrix.
 * @return A real matrix.
 * @ingroup Matrix
 */
MATRIX ConvertToReal(MATRIX m) {
	int ok = TRUE;
	MATRIX mr = newMatrix(m.nRow, m.nCol);

	if (m.mtype != COMPLEX)
		reportError(MatrixConvertToRealNotComplex, "ConvertToReal", FATAL);
	int i;
	double complex *p;
	double *q;
	double threshold = 0.1e-100;

	p = m.data;
	q = mr.data;
	for (i=0; i<m.nRow*m.nCol; i++) {
		double val;
		val = cimag(p[i]);
		if ((val > threshold) || (val < -threshold)) {
			ok = FALSE;
		} else {
			q[i] = creal(p[i]);
		}
	}
	if (!ok) {
		reportError(MatrixConvertToRealConvFail, "ConvertToReal", FATAL);
	}
	return mr;
}

/**
 * Convert a real matrix to a complex matrix.
 * @param[in] m A real matrix.
 * @return A complex matrix.
 * @ingroup Matrix
 */
MATRIX ConvertToComplex(MATRIX m) {
	MATRIX mc = newComplexMatrix(m.nRow, m.nCol);

	if (m.mtype != REAL)
		reportError(MatrixConvertToComplexNotReal, "ConvertToComplex", FATAL);
	int i;
	double complex *q;
	double *p;

	p = m.data;
	q = mc.data;
	for (i=0; i<m.nRow*m.nCol; i++) {
		q[i] = p[i] + 0.0*I;
	}
	return mc;
}

/**
 * Convert a matrix where one dimension is unity to a vector.
 * @param[in] m A matrix with either 1 row or 1 column.
 * @return The matrix as a vector equivalent.
 * @ingroup Matrix
 */
VECTOR MatrixToVector(MATRIX m) {
	VECTOR v;

	if (m.mtype == COMPLEX)
		reportError(MatrixMatrixToVectorComplexNotImpl, "MatrixToVector", FATAL);
	if ((m.nRow != 1) && (m.nCol == 1)) {
		v.n = m.nRow;
	} else if ((m.nCol != 1) && (m.nRow == 1)) {
		v.n = m.nCol;
	} else {
		reportError(MatrixMatrixToVectorNoRowOrColVec, "MatrixToVector", FATAL);
	}
	v.data = m.data;
	return v;
}

/** Convert vector to a matrix with a single column vector.
 * @param[in] v A vector.
 * @return A matrix containing the vector as a column.
 * @ingroup Matrix
 */
MATRIX VectorToMatrix_Column(VECTOR v) {
	MATRIX m;

	m.mtype = REAL;
	m.nRow = v.n;
	m.nCol = 1;
	m.data = v.data;
	return m;
}

/*
 * Extract specific vectors from a matrix.
 * Indexes is a list of row or column numbers to extract.
 * If pulling row vectors, then areRows is true.
 * NOTE:  depths are 1 based, but matrices are 0 based,
 * so subtract one from each depth value.
 * @param[in] from The source matrix from which to extract.
 * @param[in] depths A vector containing row or column indices to extract.
 * @param[in] areRows Boolean, TRUE if pulling rows, FALSE is pulling columns.
 * @return A reduced matrix containing the extracted rows or columns.
 * @ingroup Matrix
 */
MATRIX reduceMatrix(MATRIX from, VECTOR depths, int areRows) {
	MATRIX to;
	int r, c;
	int toR, toC, fromR, fromC;

	if (areRows) {
		r = depths.n;
		c = from.nCol;
	} else {
		//reportError("reduceMatrix is not implemented for pulling columns, only rows", "", FATAL);
		// extract a column vectors
		r = from.nRow;
		c = depths.n;
	}
	if (from.mtype == COMPLEX) {
		to = newComplexMatrix(r, c);
		double complex *p, *q;
		p = from.data;
		q = to.data;
		int i;
		if (areRows) {
			for (i = 0; i < depths.n; i++) {
				toR = i;
				toC = 0;
				fromR = depths.data[i] - 1.0;
				fromC = 0;
				// memcpy(to, from, size)
				memcpy(&q[ROW_MAJOR_IDX(toR, toC, from.nCol)],
						&p[ROW_MAJOR_IDX(fromR, fromC, from.nCol)], (from.nCol * sizeof(double complex)));
			}
		} else {
			// pulling columns, do one at a time
			for (c = 0; c < depths.n; c++) {
				for (i = 0; i < r; i++) {
					fromC = depths.data[c] - 1.0;
					//for (j = 0; j < c; j++) {
						q[ROW_MAJOR_IDX(i, c, to.nCol)] = p[ROW_MAJOR_IDX(i, fromC, from.nCol)];
				}
			}
		}
	} else {
		// REAL data
		//reportError("reduceMatrix is not implemented for REAL matrices", "", FATAL);
		to = newMatrix(r, c);
		double *p, *q;
		p = from.data;
		q = to.data;
		int i;
		if (areRows) {
			for (i = 0; i < depths.n; i++) {
				toR = i;
				toC = 0;
				fromR = depths.data[i] - 1.0;
				fromC = 0;
				// memcpy(to, from, size)
				memcpy(&q[ROW_MAJOR_IDX(toR, toC, from.nCol)],
						&p[ROW_MAJOR_IDX(fromR, fromC, from.nCol)], (from.nCol * sizeof(double)));
			}
		} else {
			// pulling columns, do one at a time
			for (c = 0; c < depths.n; c++) {
				for (i = 0; i < r; i++) {
					fromC = depths.data[c] - 1.0;
					//for (j = 0; j < c; j++) {
						q[ROW_MAJOR_IDX(i, c, to.nCol)] = p[ROW_MAJOR_IDX(i, fromC, from.nCol)];
				}
			}
		}
	}
	return to;
}

/**
 * Insert a vector into an array.
 * @param[in,out] to An array.
 * @param[in] from A vector of data to insert.
 * @ingroup Matrix
 */
void insertVector(double* to, VECTOR from) {
	// memcpy(to, from, size)
	memcpy(to, from.data, (from.n * sizeof(double)));
}


/**
 * Calculate A + lambda * B = C
 * @param[in] A A matrix.
 * @param[in] B A matrix.
 * @param[in] lambda A scalar value.
 * @param[in,out] C The result matrix.
 * @ingroup Matrix
 */
void APlusLambdaB(MATRIX A, double lambda, MATRIX B, MATRIX C) {
	if ( (A.nRow != B.nRow) || (A.nRow != C.nRow) ||
		(A.nCol != B.nCol) || (A.nCol != C.nCol) )
		reportError(MatrixAPlusLambdaBMatricesNotSameSize, "APlusLambdaB", FATAL);

	if ((A.mtype == REAL) && (B.mtype == REAL) && (C.mtype == REAL)) {
		// need separate pointers for each matrix based on the datatype REAL or COMPLEX
		double *a, *b, *c;

		a = A.data;
		b = B.data;
		c = C.data;
		double alpha = 1.0;
		int lda, ldc;
		lda = A.nRow;
		ldc = C.nRow;
		// memcpy(to, from, size)
		memcpy(c, b, (B.nRow*B.nCol * sizeof(double)));
		ATL_dgeadd(A.nRow, A.nCol, alpha, a, lda, lambda, c, ldc);
	} else if ((A.mtype == COMPLEX) && (B.mtype == COMPLEX) && (C.mtype == COMPLEX)) {
		// need separate pointers for each matrix based on the datatype REAL or COMPLEX
		double complex *a, *b, *c;

		a = A.data;
		b = B.data;
		c = C.data;
		double complex  alpha = 1.0;
		double complex  lambdaC = lambda;
		int lda, ldc;
		lda = A.nRow;
		ldc = C.nRow;
		// memcpy(to, from, size)
		memcpy(c, b, (B.nRow*B.nCol * sizeof(double complex)));
		ATL_zgeadd(A.nRow, A.nCol, &alpha, a, lda, &lambdaC, c, ldc);
	} else {
		reportError(MatrixAPlusLambdaBNotImplCombination, "APlusLambdaB", FATAL);
	}
}

/**
 * Calculate the ||M||_2 of a matrix, M.
 * @param[in] m A matrix.
 * @return The second norm of matrix.
 * @ingroup Matrix
 */
double complex Norm2(MATRIX m) {
	double complex result;
	result = 0.0 + 0.0*I;
	if (m.mtype == COMPLEX) {
		double complex *p;
		int i;
		p = m.data;

		for (i = 0; i < m.nCol * m.nRow; i++)
			result += p[i] * p[i];
	} else {
		reportError(MatrixNorm2NotImpl, "Norm2", ERROR);
	}
	return result;
}

/*
 *                 -1
 *  calculate F = B    V
 */
/**
 *  Calculate  \f$F=B^{-1} V\f$
 *  by solving BF = V for F then solving system of linear equations
 *  k times where k is the number of columns in V.
 * @param[in] B A matrix.
 * @param[in] V A matrix.
 * @return The resulting matrix, F.
 * @ingroup Matrix
 */
// NOT USED AT THIS TIME
// try to use linearEqSolve: MatrixInvertViaLinearEqSolve instead.
MATRIX BItimesV(MATRIX B, MATRIX V) {
	MATRIX F;	 		// F will be stored in column major order until the end
	double *VColMajor; // column major order form of V matrix
	double *FRowMajor;
	int k;
	VECTOR x,y;
	double *p;

	if ((B.mtype == COMPLEX) || (V.mtype == COMPLEX))
		reportError(MatrixComplexNotImpl, "BItimeV", FATAL);
	VColMajor = toColumnMajor(V.data, V.nRow, V.nCol);
	F = newMatrix(V.nRow, V.nCol);

	for (k = 0; k < V.nCol; k++) {

		y.data = &VColMajor[k * V.nRow];
		y.n = V.nRow;
		//x = extractVector(y);
	//VECTOR
	//linearEqSolve(MATRIX A, VECTOR v)
    //  find LU factorization of B once, then use it over and over for each k, this will be quicker than linearEqSolve
		x = linearEqSolve(B, y);
		p = F.data; // &F.data[k*F.nRow];
		insertVector(&p[k*F.nRow], x);
		td_free(x.data, x.n * sizeof(double));
	}

	FRowMajor = toRowMajor(F.data, F.nRow, F.nCol);
	freeMatrix(F);
	F.data = FRowMajor;
	td_free(VColMajor, V.nRow * V.nCol * sizeof(double));
	return F;
}

/** Compare two matrices for equality.
 * @param[in] a A matrix object.
 * @param[in] b A matrix object.
 * @return TRUE if matrix values are equal.
 */
int matrix_equal(MATRIX a, MATRIX b) {
	int equal = TRUE;

	if (a.nRow != b.nRow) return FALSE;
	if (a.nCol != b.nCol) return FALSE;
	if (a.mtype != b.mtype) return FALSE;
	if (a.mtype == REAL) {
		double *p, *q;
		int i;
		p = a.data;
		q = b.data;
		for (i=0; i<a.nCol*b.nRow; i++)
			if (p[i] != q[i]) return FALSE;
	} else {
		double complex *p, *q;
		int i;
		p = a.data;
		q = b.data;
		for (i=0; i<a.nCol*b.nRow; i++)
			if (p[i] != q[i]) return FALSE;
	}
	return equal;
}

/**
 * Split a complex matrix into real and imaginary parts.
 * @param[in] m A complex matrix.
 * @param[in,out] real A matrix containing real part.
 * @param[in,out] imag A matrix containing imaginary part.
 * @ingroup Matrix
 */
void splitMatrix(MATRIX m, MATRIX *real, MATRIX *imag) {
	int i, j;

	*real = newMatrix(m.nRow, m.nCol);
	*imag = newMatrix(m.nRow, m.nCol);

	if (m.mtype != COMPLEX)
		reportError(ComplexMatrixExpected, "splitMatrix", FATAL);
	double complex *cp;
	double *rp, *ip;
	rp = real->data;
	ip = imag->data;
	cp = m.data;
	for (i = 0; i < m.nRow; i++) {
		for (j = 0; j < m.nCol; j++) {
			*rp++ = creal(*cp);
			*ip++ = cimag(*cp);
			cp++;
		}
	}
}

/**
 * Build a complex matrix from separate real and imaginary parts.
 * @param[in,out] m A complex matrix.
 * @param[in] real A matrix containing real part.
 * @param[in] imag A matrix containing imaginary part.
 * @ingroup Matrix
 */
void mergeMatrix(MATRIX *m, MATRIX real, MATRIX imag) {
	int i, j;

	*m = newComplexMatrix(real.nRow, real.nCol);

	double complex *cp;
	double *rp, *ip;
	rp = real.data;
	ip = imag.data;
	cp = m->data;
	for (i = 0; i < real.nRow; i++) {
		for (j = 0; j < real.nCol; j++) {
			*cp = *rp + *ip*I;
			cp++;
			rp++;
			ip++;
		}
	}
}

/**
 * Set a vector to NaN.
 * @param[in,out] v A vector.
 * @ingroup Matrix
 */
void setToNaN(VECTOR v) {
	int i;
	for (i=0; i<v.n; i++) {
		v.data[i] = NAN;
	}
}

/**
 * Set a matrix to NaN.
 * @param[in,out] m A matrix.
 * @ingroup Matrix
 */
void setToNaNMatrix(MATRIX m) {
	int i;
	if (m.mtype == REAL) {
		double *p;
		p = m.data;
		for (i = 0; i < m.nRow * m.nCol; i++) {
			*p++ = NAN;
		}
	} else {
		double complex *p;
		p = m.data;
		for (i = 0; i < m.nRow * m.nCol; i++) {
			*p++ = NAN;
		}
	}
}


/**
 * Set all elements for scatterer to zero.
 * @param[in] K A MATRIX, usually Theta (a matrix defining the overlap).
 * @param[in] scatterer The scatterer of interest.  All elements (rows and columns) for this scatterer are
 *            set to zero.
 * @param[in] Nz the Nz dimension of the kernels
 * @param[in] NBeta the number of scatterers.
 * @return A copy of the Theta matrix with elements for scatterer set to zero.
 */
//I NEED TO KNOW THE Z-DIMENSION, BECAUSE THETA IS (N_z x N_beta) x (N_z x N_beta)
// K is (M x (N_z x N_beta)), so just zero rows or columns.
/* NOT CURRENTLY USED */
MATRIX zeroScatterer(MATRIX K, int scatterer, int Nz, int NBeta) {

	//print_matrix2("K", K, ROW_MAJOR);
	if (K.mtype == REAL)
		reportError("Trying to zero real data. Zero Scatterer only works on complex matrices.", "ZeroScatterer", FATAL);

	// make a copy of the matrix, or just let the compiler send a new copy of the matrix.
	MATRIX K2;
	K2 = matrix_copy(K);
	int i, m;

	double complex *p;
	//p = K2.data;
	int C;
	C = Nz * scatterer;
	int NCOL;
	NCOL = Nz * NBeta;
	p = K2.data;
	// just for testing
	p += ROW_MAJOR_IDX(0, C, NCOL);
	// find first offset, all others are just NCOL away.
	for (m = 0; m < K.nRow; m++) {
		//ROW_MAJOR_IDX(R, C, NCOL) ((R)*(NCOL) + (C))
		// rows are indexed by m (maps)
		// columns are indexed by (Nz)(scatterer)

		for (i = 0; i < Nz; i++) {
			*p++ = 0.0 + 0.0 * I;
		}
		p += (NCOL-Nz);
	}

	return K2;


}

/**
 * Multiply all elements (maps and Nz's) of scatterer by a scale factor.
 * @param[in] K The MATRIX, to scale.
 * @param[in] factor A scalar value.
 * @param[in] scatterer The scatterer of interest.  All elements (rows and columns)
 *            for this scatterer are scaled.
 * @param[in] Nz the Nz dimension of the kernels.
 * @param[in] NBeta the number of scatterers.
 * @return A new matrix with scatterer elements of K scaled.
 */
// NEED TO KNOW THE Z-DIMENSION, BECAUSE K IS (M) x (N_z x N_beta), N_z is variable
// and N_beta is constant.
// K is (M x (N_z x N_beta)), so just scale rows and columns for scatterer.
MATRIX scaleScatterer(MATRIX K, double factor, int scatterer, int Nz, int NBeta) {

	//print_matrix2("K", K, ROW_MAJOR);
	if (K.mtype == REAL)
		reportError("Trying to scale real data. scaleScatterer only works on complex matrices.", "scaleScatterer", FATAL);

	// make a copy of the matrix, or just let the compiler send a new copy of the matrix.
	MATRIX K2;
	K2 = matrix_copy(K);
	int i, m;


	double complex *p;
	int C;       // the first column to scale
	C = Nz * scatterer;
	int NCOL;    // total number of columns; isn't this just K.nCol?
	NCOL = Nz * NBeta;
	p = K2.data;
	p += ROW_MAJOR_IDX(0, C, NCOL);
	// find first offset, all others are just NCOL away.
	for (m = 0; m < K.nRow; m++) {
		//ROW_MAJOR_IDX(R, C, NCOL) ((R)*(NCOL) + (C))
		// rows are indexed by m (maps)
		// columns are indexed by (Nz)(scatterer)

		for (i = 0; i < Nz; i++) {
			*p++ *= factor;
		}
		p += (NCOL-Nz);
	}

	return K2;
}


/**
 * Multiply all  elements of non-scatterer by factor.
 * @param[in] K A MATRIX, usually K
 * @param[in] factor A scalar value
 * @param[in] scatterer The scatterer of interest.  All elements (rows and columns) for this scatterer are
 *            untouched.  All other elements are scaled by the factor.
 * @param[in] Nz the Nz dimension of the kernels
 * @param[in] NBeta the number of scatterers.
 * @return A copy of the K matrix with elements for non-scatterers scaled by the factor.
 */
//I NEED TO KNOW THE Z-DIMENSION, BECAUSE THETA IS (N_z x N_beta) x (N_z x N_beta)
// K is (M x (N_z x N_beta)), so just multiply  columns.
MATRIX scaleNonScatterer(MATRIX K, double factor, int scatterer, int Nz, int NBeta) {

	//print_matrix2("K", K, ROW_MAJOR);
	if (K.mtype == REAL)
		reportError("Trying to scale real data. Scale non-Scatterer only works on complex matrices.", "ScaleNonScatterer", FATAL);

	// make a copy of the matrix.
	MATRIX K2;
	K2 = matrix_copy(K);
	int i, m;

	double complex *p;
	//p = K2.data;
	int C;
	C = Nz * ((scatterer + 1)%NBeta);
	int NCOL;
	NCOL = Nz * NBeta;
	p = K2.data;

	if (scatterer != 1) {
		p += ROW_MAJOR_IDX(0, C, NCOL);
		// we can do two columns at a time (v_x & v_y or v_y & v_z)
		Nz = Nz *2;
		// find first offset, all others are just NCOL away.
		for (m = 0; m < K.nRow; m++) {
			//ROW_MAJOR_IDX(R, C, NCOL) ((R)*(NCOL) + (C))
			// rows are indexed by m (maps)
			// columns are indexed by (Nz)(scatterer)

			for (i = 0; i < Nz; i++) {
				*p++ *= factor;
			}
			p += (NCOL-Nz);
		}
	} else {
		// we have to skip the center column (v_y)
		// find first offset, all others are just NCOL away.
		for (m = 0; m < K.nRow; m++) {
			//ROW_MAJOR_IDX(R, C, NCOL) ((R)*(NCOL) + (C))
			// rows are indexed by m (maps)
			// columns are indexed by (Nz)(scatterer)

			// do first vx
			for (i = 0; i < Nz; i++) {
				*p++ *= factor;
			}
			//  do inbetween <vz vx>
			// move to next component column
			// actually, after the first component column, the columns are contiguous,
			// until the last component column which is a single
			// e.g. vx vy vz vx vy vz vx vy vz
			//      XX    XX XX    XX XX    XX

			// skip vy
			p += Nz;

			// do vz
			for (i = 0; i < Nz; i++) {
				*p++ *= factor;
			}
			//p += Nz;

			// do final vz
			//for (i = 0; i < Nz; i++) {
			//	*p++ *= factor;
			//}

			// move to next row
		}
	}

	//printf("after scaling ...\n");
	//print_matrix2("K", K, ROW_MAJOR);
	//print_matrix2("K2", K2, ROW_MAJOR);

	return K2;


}

/** Pad a matrix with a scalar value, i.e.
 * expand and fill the matrix with a specific value.
 * @param[in] a A matrix object.
 * @param[in] newRow Number of rows in expanded matrix.
 * @param[in] newCol Number of columns in expanded matrix.
 * @param[in] value Numeric value for empty elements.
 * @return A new matrix with padded values.
 */
MATRIX pad(MATRIX a, int newRow, int newCol, double value) {

	MATRIX bigger;

	if ((a.nRow >= newRow) && (a.nCol >= newCol))
		reportError(MatrixBiggerDim, "pad", FATAL);
	if (a.mtype == REAL) {
		bigger = newMatrix(newRow, newCol);
		double *p, *q;
		int i, r;
		p = a.data;
		q = bigger.data;
		for (r=0; r<a.nRow; r++) {
			// memcpy(to, from, size)
			memcpy(q, p, (a.nCol * sizeof(double)));
			// position pointer at end of a.row
			q += a.nCol;
			// fill remainder of row
			for (i=0; i<bigger.nCol-a.nCol; i++)
				*q++ = value;
			// advance pointer in matrix a to the next row
			p += a.nCol;
		}
		for (r=0; r<(bigger.nRow-a.nRow) * bigger.nCol; r++) {
			*q++ = value;
		}
	} else {
		reportError(MatrixComplexNotImpl, "pad", FATAL);
	}
	return bigger;
}

/** embed a matrix within a larger matrix of zeros.
 * @param[in] a A matrix object.
 * @param[in] factor Increase rows and columns by this factor.
 * @return The new embedded matrix.
 */
MATRIX embed(MATRIX a, int factor) {

	MATRIX bigger;

	if ((a.nRow != a.nCol) || ((a.nRow % 2) != 0))  {
		reportError(MatrixEmbedEvenDim, "embed", FATAL);
	}
	if (a.mtype == REAL) {
		int newR, newC, Nx;
		newR = a.nRow*factor;
		newC = a.nCol*factor;
		// create bigger matrix
		bigger = newMatrix(newR, newC);
		double *p, *q;
		int i, r;
		p = a.data;
		q = bigger.data;
		// fill matrix with zeros
		for (i=0; i<newR*newC; i++)
			*q++ = 0.0;
		// reset pointer to data
		q = bigger.data;

		// copy original data into center
		// cen=3*Nx/2+1; %% the center point (i.e. x=0)
		Nx = a.nRow;
		int cen = (newR)/2;
		// I=(cen-Nx/2):(cen+Nx/2-1);
		int ColStart = cen - (Nx/2);
		int RowStart = cen - (Nx/2);
		q += ROW_MAJOR_IDX(RowStart, ColStart, newC);
		for (r=0; r<a.nRow; r++) {
			// memcpy(to, from, size)
			memcpy(q, p, (a.nCol * sizeof(double)));
			// position pointer at start of next row in bigger matrix
			q += newC;
			// advance pointer in matrix a to the next row
			p += a.nCol;
		}
		//for (r=0; r<(bigger.nRow-a.nRow) * bigger.nCol; r++) {
		//	*q++ = value;
		//}
	} else {
		int newR, newC, Nx;
		newR = a.nRow*factor;
		newC = a.nCol*factor;
		// create bigger matrix
		bigger = newComplexMatrix(newR, newC);
		double complex *p, *q;
		int i, r;
		p = a.data;
		q = bigger.data;
		// fill matrix with zeros
		for (i=0; i<newR*newC; i++)
			*q++ = 0.0;
		// reset pointer to data
		q = bigger.data;

		// copy original data into center
		// cen=3*Nx/2+1; %% the center point (i.e. x=0)
		Nx = a.nRow;
		int cen = (newR)/2;
		// I=(cen-Nx/2):(cen+Nx/2-1);
		int ColStart = cen - (Nx/2);
		int RowStart = cen - (Nx/2);
		q += ROW_MAJOR_IDX(RowStart, ColStart, newC);
		for (r=0; r<a.nRow; r++) {
			// memcpy(to, from, size)
			memcpy(q, p, (a.nCol * sizeof(double complex)));
			// position pointer at start of next row in bigger matrix
			q += newC;
			// advance pointer in matrix a to the next row
			p += a.nCol;
		}
	}
	return bigger;
}

/** Read a vector from a file
 * Format of the file can be eol, tab, comma, space, separated list of numbers
 * @param[in] inputFilename Name of file to read.
 * @return Vector object containing data from file.
 * @ingroup Matrix
 */
VECTOR readText(char *inputFilename) {
	VECTOR v;
	FILE *fp;
	double dummy;
	int conv;

	fp = fopen(inputFilename, "r");
	if (fp != 0) {
		// scan file first to determine the how many numbers
		int count = 0;
		while ((conv = fscanf(fp, "%lf", &dummy)) != EOF) {
			count += 1;
		}
		// allocate space for the vector
		v = newVector(count);

		// reset the pointer;
		rewind(fp);
		// read and insert the numbers into the vector
		count = 0;
		while (!feof(fp) && count < v.n) {
			conv = fscanf(fp, "%lf", &v.data[count]);
			count += 1;
		}
		fclose(fp);
	} else {
		reportError(FileOpenError, inputFilename, FATAL);
	}
	return v;
}

/* Auxiliary routine: printing a matrix */
/**
 * Print a matrix.
 * @param[in] desc A description or name of the matrix.
 * @param[in] m The number of rows in the matrix.
 * @param[in] n The number of columns in the matrix.
 * @param[in] a An array of the data in the matrix.
 * @param[in] lda Leading dimension of the matrix.
 * @param[in] order Layout of the matrix data, either row-major or column-major.
 * @ingroup Matrix
 */
void print_matrix(char* desc, int m, int n, double* a, int lda, enum matrixOrder order) {
	int i, j;
	printf("\n %s\n", desc);
	for (i = 0; i < m; i++) {
		for (j = 0; j < n; j++)
			if (order == ROW_MAJOR) {
				printf(" %6.2f", a[i* lda + j]);
			} else { // column major order
				printf(" %6.2f", a[i + j * lda]);
			}
		printf("\n");
	}
}

/**
 * Print a matrix.
 * @param[in] desc A description or name of the matrix.
 * @param[in] m A matrix.
 * @param[in] order Layout of the matrix data, either row-major or column-major.
 * @ingroup Matrix
 */
void print_matrix2(char* desc, MATRIX m, enum matrixOrder order) {
	int i, j;
	int idx;

	printf("\n %s= [\n", desc);
	for (i = 0; i < m.nRow; i++) {
		for (j = 0; j < m.nCol; j++) {
			if (order == ROW_MAJOR) {
				idx = i* m.nCol + j;
			} else { // column major order
				idx = i + j * m.nRow;
			}
			if (m.mtype == REAL) {
				double *p;
				p = m.data;
				printf(" %25.12e,", p[idx]);
			} else {
				double complex *p;
				p = m.data;
				printf(" %25.12e+%25.12ei,", creal(p[idx]), cimag(p[idx]));
			}
		}
		printf(";\n");
	}
	printf("]\n");
}

/**
 * Print a matrix to a file.
 * @param[in] desc A description or name of the matrix.  This is also the name
 * 					of the file.
 * @param[in] m A matrix.
 * @param[in] order Layout of the matrix data, either row-major or column-major.
 * @ingroup Matrix
 */
void print_matrix2toFile(char* desc, MATRIX m, enum matrixOrder order) {
	int i, j;
	int idx;
	FILE *fp;

	fp = fopen(desc,"w");
	fprintf(fp,"\n %s = [ \n", desc);
	for (i = 0; i < m.nRow; i++) {
		for (j = 0; j < m.nCol; j++) {
			if (order == ROW_MAJOR) {
				idx = i* m.nCol + j;
			} else { // column major order
				idx = i + j * m.nRow;
			}
			if (m.mtype == REAL) {
				double *p;
				p = m.data;
				//fprintf(fp," %6.2f,", p[idx]);
				fprintf(fp," %e,", p[idx]);
			} else {
				double complex *p;
				p = m.data;
				fprintf(fp," %25.12e + %25.12ei, ", creal(p[idx]), cimag(p[idx]));
			}
		}
		fprintf(fp,";\n");
	}
	fprintf(fp,"]\n");
	fclose(fp);
}

