/*
 * matrix.h
 *
 *  Created on: Apr 22, 2010
 *      Author: Brenda Javornik (e-mail: brendaj (at) cora.nwra.com)
 */

#ifndef MATRIX_H_
#define MATRIX_H_

#include "objectfits.h"
#include "utilities.h"
#include <complex.h>


/**
 * Type of data kept in matrix: either Real (double) or complex (double complex)
 * @ingroup Matrix
 */
typedef enum {REAL, COMPLEX} MatrixType;

/**
 * MATRIX is a main datatype and contains either double or double complex
 * data.
 * @ingroup Matrix
 */
typedef struct {
  void *data;			///< array of data values stored in row-major order
  int nRow;				///< number of rows in matrix
  int nCol;				///< number of columns in matrix
  MatrixType mtype;		///< type of data, either real or complex.
} MATRIX;

/**
 * VECTOR is a main datatype and contains data of type double only.
 * @ingroup Matrix
 */
typedef struct {
  double *data;	///< array of data values
  int n;  		///< number of elements in vector
} VECTOR;

/**
 * MatrixOrder designates the layout of data in a matrix.
 * @ingroup Matrix
 */
enum matrixOrder {ROW_MAJOR, COL_MAJOR};

MATRIX newMatrix(int r, int c);
MATRIX newMatrixI(int r, int c, double *Idata);

MATRIX newComplexMatrix(int r, int c);

void freeMatrix(MATRIX m);

VECTOR newVector(int n);
void freeVector(VECTOR v);

unsigned int setRandomSeedToTime();
unsigned int setRandomSeed(unsigned int seed);

MATRIX newRandomMatrix(int r, int c);

VECTOR multiplyMV(MATRIX m, VECTOR v);

// MATRIX multiplyMM(MATRIX m1, MATRIX m2); no longer used
MATRIX multiplyMM2(char TransA, char TransB, MATRIX A, MATRIX B);
MATRIX multiplyMM2NoForceC(char TransA, char TransB, MATRIX A, MATRIX B);
MATRIX multiplyMMFull(char TransA, char TransB, double alpha, double beta, MATRIX A, MATRIX B);

/**
 * Multiply a matrix, m, by a diagonal matrix represented by its diagonal elements in
 * vector, d
 */
MATRIX multiplyMD(MATRIX m, VECTOR d);

// multiply a diagonal matrix, whose diagonal elements are contained in the
// vector d, by a matrix m
MATRIX multiplyDM(VECTOR d, MATRIX m);

VECTOR multiplyDV(VECTOR d, VECTOR v);

MATRIX multiplyScalar(MATRIX m, double scalar);
MATRIX multiplyScalarZ(MATRIX m, double complex scalar);
MATRIX starMultiply(MATRIX P, MATRIX Q);

MATRIX multiplyMMplusC(MATRIX A, MATRIX B, double beta, MATRIX C);
MATRIX multiplyMMplusCZ(MATRIX A, MATRIX B, double beta, MATRIX C);
MATRIX multiplyABplusC(char TransA, char TransB, MATRIX A, MATRIX B, double beta, MATRIX C);
MATRIX multiplyAlphaABplusC(char TransA, char TransB, MATRIX A, MATRIX B, double alpha, double beta, MATRIX C);


MATRIX toMatrix(fits2d f);
VECTOR toVector(fits2d f);

fits2d toFitsObject(VECTOR v);
fits2d MatrixToFitsObject(MATRIX m);
fits3d MatrixToFits3dObject(MATRIX m);

double* toRowMajor(double* A, int R, int C);
double* toColumnMajor(double* A, int R, int C);

MATRIX toRowMajor2(MATRIX m);
MATRIX toColumnMajor2(MATRIX m);

MATRIX expandDiagonalMatrix(VECTOR v);

MATRIX transpose(MATRIX m);

MATRIX conjugate(MATRIX m);

MATRIX inverseD(MATRIX m);

MATRIX inverse(MATRIX m);

MATRIX Cholesky(MATRIX m);

MATRIX ConvertToReal(MATRIX m);

MATRIX ConvertToComplex(MATRIX m);

VECTOR MatrixToVector(MATRIX m);

MATRIX VectorToMatrix_Column(VECTOR v);

MATRIX reduceMatrix(MATRIX from, VECTOR depths, int areRows) ;

void insertVector(double* to, VECTOR from);

void APlusLambdaB(MATRIX A, double lambda, MATRIX B, MATRIX C);

double complex Norm2(MATRIX m);

MATRIX BItimesV(MATRIX B, MATRIX V);

int matrix_equal(MATRIX a, MATRIX b);

void setToNaN(VECTOR v);
void setToNaNMatrix(MATRIX m);

// MATRIX ZeroScatterer(MATRIX K, int scatterer, int Nz, int NBeta);

MATRIX pad(MATRIX a, int newRow, int newCol, double value);
MATRIX embed(MATRIX a, int factor);
void splitMatrix(MATRIX m, MATRIX *real, MATRIX *imag);
void mergeMatrix(MATRIX *m, MATRIX real, MATRIX imag);

MATRIX zeroScatterer(MATRIX K, int scatterer, int Nz, int NBeta);
MATRIX scaleScatterer(MATRIX K, double factor, int scatterer, int Nz, int NBeta);
MATRIX scaleNonScatterer(MATRIX K, double factor, int scatterer, int Nz, int NBeta);

/** read a vector from a text file (ASCII file)
 *
 */
VECTOR readText(char *filename);

void print_matrix( char* desc, int m, int n, double* a, int lda, enum matrixOrder order);
void print_matrix2(char* desc, MATRIX m, enum matrixOrder order);
void print_matrix2toFile(char* desc, MATRIX m, enum matrixOrder order);


int testTranspose();

#endif /* MATRIX_H_ */
