/*
 * objectfits.h
 *
 *  Created on: Apr 16, 2010
 *      Author: Brenda Javornik (e-mail: brendaj (at) cora.nwra.com)
 */

#ifndef OBJECTFITS_H_
#define OBJECTFITS_H_

#include <fitsio.h>
#include "FITSParameters.h"

/**
 * @defgroup FITS FITS - FITS handles input and output of FITS files.
 * @ingroup SupportingModules
 */

/**
 * A 2-d array of double values.  Data read from or written to .fits files are
 * kept in these buffers.
 * @ingroup FITS
 */
typedef struct {
  double *data;	///< numeric values
  int xDim;		///< number of rows
  int yDim;		///< number of columns
} fits2d;

/**
 * A 2-d array of complex values.  The third dimension is size 2 -- one for Real data
 * the other for Imaginary data.  Fits3d is a buffer for writing complex data to a .fits
 * file.
 * @ingroup FITS
 */
typedef struct {
  double **Real;		///< Real values
  double **Imaginary;	///< Imaginary values
  int xDim;				///< number of rows
  int yDim;				///< number of columns
  int zDim;				///< set to 2 for Real and Imaginary parts
} fits3d;

/**
 * Keeps indices for a location in a 3-d .fits file
 * @ingroup FITS
 */
typedef struct {
	int x;	///< row number
	int y;	///< column number
	int z;	///< xy-plane number
	int t;  ///< 4th dimension of FITS file
} location;

/**
 * File information for a 3-d .fits file;  This information is kept and updated
 * during incremental reading and writing to multidimensional .fits files.
 * @ingroup FITS
 */
typedef struct {
	fitsfile *fptr;		///< file pointer to a multidimensional .fits file
	location current;	///< current location in the .fits file for reading or writing
	location max;		///< last position of data in .fits file or maximum dimensions for .fits file
	int bitpix;  		/**< datatype of the image:
							BYTE_IMG   =   8   ( 8-bit byte pixels, 0 - 255)
							SHORT_IMG  =  16   (16 bit integer pixels)
							LONG_IMG   =  32   (32-bit integer pixels)
							FLOAT_IMG  = -32   (32-bit floating point pixels)
							DOUBLE_IMG = -64   (64-bit floating point pixels)
						 */

	int dimension;		///< number of dimensions for this .fits file
} MultiDimFits;

/**
 * Which plane of multi-dimensional fits file to read or write.
 */
typedef struct {
	int z;
	int t;
} WhichPlane;

// The operations on the objectfits datatype
fits2d readfits(char *filename);
int writefits(char *filename, fits3d data);
int writefits2d(char *filename, fits2d data);
int writefits2dWithLambda(char *filename, fits2d data, double lambda);

MultiDimFits openMultiDimFitsRead(char *filename);
void openMultiDimFitsWrite(char *filename, MultiDimFits *fileInfoPtr, FITSParameters headerParameters);
fits2d readNextPlane(MultiDimFits fileInfo, WhichPlane whichPlane);
void writeNextPlane(MultiDimFits fileInfo, WhichPlane whichPlane, fits2d data);
int closeMultiDimFits(MultiDimFits fileInfo);

#endif /* OBJECTFITS_H_ */
