/*
 * readfits.c
 *
 *  Created on: Apr 16, 2010
 *      Author: Brenda Javornik (e-mail: brendaj (at) cora.nwra.com)
 */

#include <string.h>
#include <stdio.h>
#include <math.h>
#include <complex.h>
#include <fftw3.h>
#include <fitsio.h>
#include "objectfits.h"
#include "utilities.h"
#include "FITSParameters.h"
#include "errorHandler.h"
#include "MatrixFunctionErrorMessages.h"

#define DEBUG 0

#define MAXDIM 4

/**
 * Read a 2D or 1D FITS file.
 * @param[in] filename Name of FITS file to open and read.
 * @ingroup FITS
 */
fits2d readfits(char *filename) {

	fitsfile *fptr; /* FITS file pointer, defined in fitsio.h */
	int status = 0; /* CFITSIO status value MUST be initialized to zero! */
	int bitpix, naxis, ii;
	long naxes[2] = { 1, 1 }, fpixel[2] = { 1, 1 };
	double *pixels = NULL;
	fits2d fitsContent;
	char format[20], hdformat[20];
	int y;

	if (!fits_open_file(&fptr, filename, READONLY, &status)) {
		if (!fits_get_img_param(fptr, 2, &bitpix, &naxis, naxes, &status)) {
			if (naxis > 2 || naxis == 0)
				reportError(UnsupportedFitsFormat, "readfits.pixels", FATAL);
			else {
				/* get memory for data matrix */
				pixels = (double *) td_malloc(naxes[0] * naxes[1] * sizeof(double));
				if (pixels == NULL)
					reportError(OutOfMemory, "readfits.pixels", FATAL);
			}

			if (DEBUG) {
				if (bitpix > 0) { /* set the default output format string */
					strcpy(hdformat, " %7d");
					strcpy(format, " %7.0f");
				} else {
					strcpy(hdformat, " %15d,");
					strcpy(format, " %15.5f,");
				}

				printf("\n      "); /* print column header */
				for (ii = 1; ii <= naxes[0]; ii++)
					printf(hdformat, ii);
				printf("\n"); /* terminate header line */
			}

			int idx;
			// loop over all the rows in the image, top to bottom
			for (fpixel[1] = naxes[1]; fpixel[1] >= 1; fpixel[1]--) {
				idx = (fpixel[1] - 1) * naxes[0];
				//  fits_read_pix(fptr, type,    fpixel, nelements ...)
				if (fits_read_pix(fptr, TDOUBLE, fpixel, naxes[0], NULL,
						&pixels[idx], NULL, &status)) // read row of pixels
					break; // jump out of loop on error

				ii = fpixel[1] - 1;

				if (DEBUG) {
					printf(" %4ld, ", fpixel[1]); // print row number
					for (y = 0; y < naxes[0]; y++) {
						printf(format, pixels[ii * naxes[0] + y]); // print each value
					}
					printf("\n"); // terminate line
				}

			}
		}

		fits_close_file(fptr, &status);
	} else {
		reportError(FileOpenError, filename, FATAL);
	}

	fitsContent.data = pixels;
	fitsContent.xDim = naxes[1];
	fitsContent.yDim = naxes[0];
	return fitsContent;

}

/**
 * 	 Open a FITS file for reading; return file info in object containing
	 fitsfile *fptr;
	 current location x, y, z
	 @param[in] filename Name of the file to open for reading.
	 @return MultiDimFits object with
	 fitsfile *fptr;
	 current location x, y, z
	 @ingroup FITS
 */
MultiDimFits openMultiDimFitsRead(char *filename) {

	fitsfile *fptr; /* FITS file pointer, defined in fitsio.h */
	MultiDimFits fileInfo;
	int status = 0; /* CFITSIO status value MUST be initialized to zero! */
	int bitpix, naxis;
	long naxes[MAXDIM] = { 1, 1, 1, 1 };
	int error = TRUE;

	if (!fits_open_file(&fptr, filename, READONLY, &status)) {
		if (!fits_get_img_param(fptr, MAXDIM, &bitpix, &naxis, naxes, &status)) {
			if (naxis > MAXDIM)
				reportErrorInt(ReadFitsFileBadDimensions, naxis, "openMultiDimFitsRead", ERROR);
			else {
				error = FALSE;
				fileInfo.fptr = fptr;
				fileInfo.current.x = 0;
				fileInfo.current.y = 0;
				fileInfo.current.z = 0;
				fileInfo.dimension = naxis;
				fileInfo.max.x = naxes[1]; // perhaps, this is why the input data needs to be transposed
				fileInfo.max.y = naxes[0]; // notice, x and y are switched.
				if (naxis > 2)
					fileInfo.max.z = naxes[2];
				else
					fileInfo.max.z = 1;
				if (naxis > 3)
					fileInfo.max.t = naxes[3];
				else
					fileInfo.max.t = 1;
				fileInfo.bitpix = bitpix;
			}
		}
	}

	if (error) {
		fits_report_error(stderr, status);
		reportError(FileOpenError, filename, FATAL);
		fileInfo.fptr = NULL;
		fileInfo.current.x = -1;
		fileInfo.current.y = -1;
		fileInfo.current.z = -1;
		fileInfo.current.t = -1;
		fileInfo.max.x = -1;
		fileInfo.max.y = -1;
		fileInfo.max.z = -1;
		fileInfo.max.t = -1;
	}
	return fileInfo;
}

/**
 * Open a 3D FITS file for writing; Write the header information including the
 * lambda value used in data calculations. Returns file info in object containing:
 * fitsfile *fptr;
 * current location x, y, z
 *
 * @param[in] filename Name of the file to open for writing.
 * @param[in,out] fileInfoPtr Object containing information about FITS file.
 * @param[in] headerParameters List of parameter names, values and comments to write in FITS header.
 * @ingroup FITS
 */
void openMultiDimFitsWrite(char *filename, MultiDimFits *fileInfoPtr, FITSParameters headerParameters) {

	fitsfile *fptr; /* FITS file pointer, defined in fitsio.h */
	int status = 0; /* CFITSIO status value MUST be initialized to zero! */
	int naxis;
	long naxes[MAXDIM] = { 1, 1, 1, 1 };
	int error = TRUE;
	MultiDimFits fileInfo;

	fileInfo = *fileInfoPtr;

	char overwriteFlag = '!';
	char *tempFilename;

	tempFilename = (char *) td_malloc((strlen(filename) + 2) * sizeof(char));
	strcpy(&tempFilename[1], filename);
	tempFilename[0] = overwriteFlag;

	if (DEBUG) printf("filename = %s\n", filename);

	status = 0; /* initialize status before calling fitsio routines */
	if (!fits_create_file(&fptr, tempFilename, &status)) { /* create new  or overwrite existing one file */

		fileInfoPtr->fptr = fptr;
		naxis = fileInfo.dimension;
		naxes[0] = fileInfo.max.y; // NOTICE:  yDim;
		naxes[1] = fileInfo.max.x; // xDim;
		naxes[2] = fileInfo.max.z; // zDim;
		naxes[3] = fileInfo.max.t; // tDim;

		/* Create the primary array image (64-bit double precision floating point */
		fits_create_img(fptr, DOUBLE_IMG, naxis, naxes, &status);

		// write parameters
		if (headerParameters.n > 0) {

		    char card[MAX_CARD_LENGTH_FOR_FITS];
			// for numeric scientific notation
			int i;
			for (i=0; i<headerParameters.n; i++) {
				snprintf(card, MAX_CARD_LENGTH_FOR_FITS, "%-7s = %20s / %s",
						headerParameters.names[i], headerParameters.scientificNotation[i], headerParameters.comments[i]);
				fits_write_record(fptr, card, &status);
				if (status != 0) {
					fits_report_error(stderr, status); // print out any error messages
					reportError(WriteFitsHeaderError, "openMultiDimFitsWrite", FATAL);
					fits_close_file(fptr, &status); // close the file
					error = TRUE;
				} else {
					error = FALSE;
				}
			}

		} else {
			error = FALSE;
		}

	}

	td_free(tempFilename, (strlen(filename) + 2) * sizeof(char));

	if (error) {
		fits_report_error(stderr, status);
		reportError(FileOpenError, filename, ERROR);
		fileInfo.fptr = NULL;
		fileInfo.current.x = -1;
		fileInfo.current.y = -1;
		fileInfo.current.z = -1;
		fileInfo.current.t = -1;
		fileInfo.max.x = -1;
		fileInfo.max.y = -1;
		fileInfo.max.z = -1;
		fileInfo.max.t = -1;
	}

}

/**
 * Read the next plane of a 3D FITS file.
 * @param[in] fileInfo Object containing FITS file info, such as file ptr, dimensions, etc.
 * @param[in] whichPlane Which plane (z-dimension) for the xy-plane to write.
 * @return A fits2d object containing the data of the xy-plane for the plane specified.
 * @ingroup FITS
 */
fits2d readNextPlane(MultiDimFits fileInfo, WhichPlane whichPlane) {
	// keep file pointer around
	// read next plane from fits file
	// set current location
	// return fits2d

	int status = 0; /* CFITSIO status value MUST be initialized to zero! */
	int ii;
	long fpixel[MAXDIM] = { 1, 1, 1, 1 };

	// Keep this static to reuse the fftwplan;  This requires all input fits files to have the same dimension
	static double *pixels = NULL;
	static int pixelsInit = FALSE;
	static int xDim = 0, yDim = 0;

	char format[20], hdformat[20];
	int idx;
	fits2d fitsPlane;

	if (fileInfo.fptr != NULL) {
		if ((!pixelsInit) || (xDim != fileInfo.max.x) || (yDim != fileInfo.max.y)) {
			if (pixels != NULL) fftw_free(pixels); // fftw_free(pixels);
			pixels = (double *) fftw_malloc(fileInfo.max.x * fileInfo.max.y * sizeof(double));
			if (pixels == NULL)
				reportError(OutOfMemory, "readfits.pixels", FATAL);
			pixelsInit = TRUE;
			xDim = fileInfo.max.x;
			yDim = fileInfo.max.y;
		}
		if (DEBUG) {
			if (fileInfo.bitpix > 0) { /* set the default output format string */
				strcpy(hdformat, " %7d");
				strcpy(format, " %7.0f");
			} else {
				strcpy(hdformat, " %15d,");
				strcpy(format, " %10.5f,");
			}

			printf("\n      "); /* print column header */
			for (ii = 1; ii <= fileInfo.max.y; ii++)
				printf(hdformat, ii);
			printf("\n"); /* terminate header line */
		}
		// update which plane we are retrieving
		if ((fileInfo.dimension > 2) && (whichPlane.z > fileInfo.max.z))
				reportErrorInt(ReadWriteFitsBadZDimension, whichPlane.z, "readNextPlane", FATAL);
		if ((fileInfo.dimension > 3) && (whichPlane.t > fileInfo.max.t))
				reportErrorInt(ReadWriteFitsBadTDimension, whichPlane.t, "readNextPlane", FATAL);
		else {
			// set fpixel[2] and fpixel[3] = plane to retrieve
			fpixel[3] = whichPlane.t;
			fpixel[2] = whichPlane.z; //fileInfo.current.z;
			fpixel[0] = 1;
			/* loop over all the rows in the image, top to bottom */
			for (fpixel[1] = fileInfo.max.x; fpixel[1] >= 1; fpixel[1]--) {

				idx = (fpixel[1] - 1) * fileInfo.max.y;
				//  fits_read_pix(fptr, type,    fpixel, nelements ...)
				if (fits_read_pix(fileInfo.fptr, TDOUBLE, fpixel, fileInfo.max.y,
						NULL, &pixels[idx], NULL, &status)) // read row of pixels
					break; // jump out of loop on error
			}
		}
	}

	if (status)
		fits_report_error(stderr, status); /* print any error message */

	fitsPlane.data = pixels;
	fitsPlane.xDim = fileInfo.max.x;
	fitsPlane.yDim = fileInfo.max.y;

	return fitsPlane;

}

/**
 * Write the next plane of a 3D FITS file.
 * @param[in] fileInfo Object containing FITS file info, such as file ptr, dimensions, etc.
 * @param[in] whichPlane Which plane (z-dimension) for the xy-plane to write.
 * @param[in] data A fits2d object containing the data of an xy-plane.
 * @ingroup FITS
 */
void writeNextPlane(MultiDimFits fileInfo, WhichPlane whichPlane, fits2d data) {
	int status = 0; /* CFITSIO status value MUST be initialized to zero! */
	long fpixel[MAXDIM] = { 1, 1, 1, 1 };

	// keep file pointer around
	// write next plane to fits file
	// update current location
	// return status

	// validate whichPlane; reportError as needed
	if ((whichPlane.z > fileInfo.max.z) || (whichPlane.z < 1)) {
		reportErrorInt(ReadWriteFitsBadZDimension, whichPlane.z, "writeNextPlane", FATAL);
	}
	if (fileInfo.dimension == 4) {
		if ((whichPlane.t > fileInfo.max.t) || (whichPlane.t < 1))
			reportErrorInt(ReadWriteFitsBadTDimension, whichPlane.t, "writeNextPlane", FATAL);
	}

	if (fileInfo.fptr != NULL) {
		double *thedata;
		thedata = data.data;
		// Write the array of doubles to the image
		fpixel[0] = 1; // this is the starting point for output
	    // move to next frame; set fpixel to next value in z dimension
	    fpixel[2] = whichPlane.z;
	    fpixel[3] = whichPlane.t;
		int idx;
		// loop over all the rows in the image, top to bottom
		for (fpixel[1] = fileInfo.max.x; fpixel[1] >= 1; fpixel[1]--) {
			idx = (fpixel[1] - 1) * fileInfo.max.y;
			fits_write_pix(fileInfo.fptr, TDOUBLE, fpixel, fileInfo.max.y, &thedata[idx],
					&status);
			if (status)
				fits_report_error(stderr, status); /* print any error message */
		}
	} else {
		// error Null filepointer; must open file before writing to it.
		reportError(WriteFitsFileNotOpen, "writeNextPlane", FATAL);
	}
}

/**
 * Close 3D FITS file
 * @param[in] fileInfo Object containing FITS file info, such as file ptr, dimensions, etc.
 * @ingroup FITS
 */
int closeMultiDimFits(MultiDimFits fileInfo) {
	int status = 0;   /* CFITSIO status value MUST be initialized to zero! */

	fits_close_file(fileInfo.fptr, &status);
	if (status)
		fits_report_error(stderr, status);
	return (status);
}

