/*
 * Model.c
 *
 *  Created on: Aug 18, 2010
 *      Author: Brenda Javornik (e-mail: brendaj (at) cora.nwra.com)
 */

#include <math.h>
#include "Constants.h"
#include "Model.h"
#include "BaseDirectory.h"
#include "InversionErrorMessages.h"
#include "errorHandler.h"
#include "matrix.h"
#include "MapDepthMatrix.h"
#include "FITSParameters.h"
#include "utilities.h"
#include "averagingKernel.h"
#include "ProblemDimension.h"
#include "IgnoreK0.h"

#define DEBUG 0
#define DEBUG1 0

/**
 * @defgroup Model Model - Inversion Solution and accompanying data.
 * @ingroup InversionData
 */

/** Initialize Model object, allocate space for data and constant values.
 * Use for OLA inversions.
 * @param[in,out] model Initialize this Model.
 * @param[in] X Number of grid points, Nx
 * @param[in] Y Number of grid points, Ny
 * @param[in] M Number of measurements
 * @param[in] Z Number of grid points, Nz
 */
void initializeModel(Model *model, int X, int Y, int M, int Z) {

	BOOLEAN shift = FALSE;
	model->w = newMapDepthMatrix(X, Y, 1, M, shift);
	model->parms.n = MAX_FITS_PARMS;
	model->parms.names[0] = Constants.ModelMuName;
	model->parms.names[1] = Constants.ModelSigmaName;
	model->parms.names[2] = Constants.ModelNuName;
	model->parms.comments[0] = Constants.ModelComment;
	model->parms.comments[1] = Constants.ModelComment;
	model->parms.comments[2] = Constants.ModelComment;
	model->q.data = NULL;
	model->AvgKern.data = NULL;
	model->noise.data = NULL;
}

/**
 * Initialize output model for use by RLS inversion
 * @param[in,out] model Pointer to Model structure
 * @param[in] kernelTdm Matrix containing 3D RLS Kernel.
 */
void initializeOutputModel(Model *model, MAP_DEPTH_MATRIX kernelTdm) {
	int x, y, z, t, v;
	x = kernelTdm.nX;
	y = kernelTdm.nY;
	z = kernelTdm.nZ;
	int useShift = TRUE;

	v = 1;
	if (getProblemDimension() == PDC.D5) {
		v = 3;
	}
	t = 1;
	// output matrices are based on the kernel dimensions
	// allocate space
	model->bFull = newMapDepthMatrixV(x, y, z, t, v, useShift); // t = 1
	t = kernelTdm.nT;
	z = 1;
	model->dTilde = newMapDepthMatrix(x, y, z, t, useShift); // z = 1
	z = kernelTdm.nZ;
	if (calcAvgKernelCov()) {
		t = getNumberOfAveragingKernels();
		model->AvgKern = newMapDepthMatrixV(x, y, z, t * v, v, useShift); // t = # of averaging kernels to calculate
		model->noise = newMapDepthMatrixV(x, y, z, t * v, v, useShift);
	}
	model->kxGridValues = newMatrix(x, y);
	model->kyGridValues = newMatrix(x, y);
	model->lambdaGridValues = newMatrix(x, y);
	model->kxkyDone = FALSE;
	model->parms.n = MAX_FITS_PARMS - 1;  // there are only two parameters for RLS inversions
	model->parms.names[0] = Constants.ModelLambda1Name;
	model->parms.names[1] = Constants.ModelLambda2Name;
	model->parms.comments[0] = Constants.ModelLambdaComment;
	model->parms.comments[1] = Constants.ModelLambdaComment;
}

/**
 * Initialize output model for use by forward calculation
 * @param[in,out] model Pointer to Model structure
 * @param[in] kernelTdm Matrix containing 3D Kernel.
 */
void initializeForwardOutputModel(Model *model, MAP_DEPTH_MATRIX kernelTdm) {
	int x, y, z, t;
	x = kernelTdm.nX;
	y = kernelTdm.nY;
	z = kernelTdm.nZ;
	int useShift = TRUE;

	t = 1;
	// output matrices are based on the kernel dimensions
	// allocate space
	//model->bFull = newMapDepthMatrixV(x, y, z, t, v, useShift); // t = 1
	t = kernelTdm.nT;
	z = 1;
	// dTilde is in real space, but kernels are in wave space, so
	// make the conversions in the dimensions
	y = 2*(y-1);
	model->dTilde = newMapDepthMatrix(x, y, z, t, useShift); // z = 1
	z = kernelTdm.nZ;

	model->kxkyDone = FALSE;
	model->parms.n = 1;
	model->parms.names[0] = Constants.ModelNoiseScaleName;
	model->parms.comments[0] = Constants.NoiseScaleComment;
}

/**
 * Free all memory associated with outputModel.
 * @ingroup OutputModel
 */
void freeModel(Model *model) {
	destroyMapDepthMatrix(model->AvgKern);
	destroyMapDepthMatrix(model->noise);
	destroyMapDepthMatrix(model->q);
	destroyMapDepthMatrix(model->w);

}

/**
 * Record parameter values in the model.
 * @param[in,out] model A Model object.
 * @param[in] mu Value for mu parameter.
 * @param[in] sigma Value for sigma parameter.
 */
void setParameters(Model *model, double mu, double sigma) {
	snprintf(&(model->parms.scientificNotation[0][0]), MAX_FITS_PARM_VALUE, FITSParametersConstants.sciFiFormat, mu);
	snprintf(&(model->parms.scientificNotation[1][0]), MAX_FITS_PARM_VALUE, FITSParametersConstants.sciFiFormat, sigma);
	model->mu = mu;
	model->sigma = sigma;
	model->parms.n = 2;
}

/**
 * Record multiple parameter values in the model.
 * @param[in,out] model A Model object.
 * @param[in] values Pointer to array of parameter values. NOTE: there is a special ordering
 * to the parameters, mu, sigma, then nu (which is optional).
 * @param[in] nValues How many parameter values.
 */
void setParameters2(Model *model, double *values, int nValues) {
	int i;
	for (i=0; i<nValues; i++) {
		snprintf(&(model->parms.scientificNotation[i][0]), MAX_FITS_PARM_VALUE, FITSParametersConstants.sciFiFormat, values[i]);
	}
	model->mu = values[0];
	model->sigma = values[1];
	if (nValues > 2) {
		model->nu = values[2];
	} else {
		model->nu = 0.0;  // if no crosstalk, then nValues = 2, and
						  // set default value for nu (equal to zero), when there is no crosstalk
		                  // This is related to Parameter.initLoopParameters() when there is no
		                  // crosstalk, we need a default value for nu.
	}
	model->parms.n = nValues;
}

/**
 * Clear parameter values (mu and sigma) in the model.
 * Output FITS files will not have sigma and mu parameters in the header.
 * @param[in,out] model A Model object.
 */
void clearParameters(Model *model) {
	model->parms.n = 0;
}

/**
 * Save inversion weights, w, for wave number, kx, ky.
 * @param[in,out] model A Model object.
 * @param[in] w Inversion weights.
 * @param[in] kx X-index in Fourier space.
 * @param[in] ky Y-index in Fourier space.
 */
void saveW(Model *model, MATRIX w, int kx, int ky) {
	if ((w.nRow != model->w.nT) || (w.nCol != model->w.nZ)) {
		reportError(ModelBadDimForW, "Model.saveW", FATAL);
	}
	insertMapDepth(kx, ky, w, &(model->w));
}

/**
 * Save inversion solution, b, for RLS 1D.
 * @param[in,out] model A Model object.
 * @param[in] b Solution matrix.
 * @param[in] nStep Parameter step number.
 */
void saveB(Model *model, MATRIX b, int nStep) {

	double *p;
	p = model->bVECTORS.data;

	// insert solution vectors as rows, then just transpose to output as column vectors
	VECTOR bv;
	bv = MatrixToVector(b);
	int idx;
	idx = (nStep - 1) * bv.n;
	insertVector(&p[idx], bv);
}

/**
 * Calculate q, Averaging Kernel and noise squared.
 * @param[in] model A Model object.
 * @param[in] K Kernel matrix
 * @param[in] tau Travel-map matrix
 * @param[in] Lambda Travel-map noise covariance matrix
 * @param[in] DValueShift Boolean indicating whether to shift output matrices q, and averaging kernel, and noise.
 */
void calculateDerivedValues(Model *model, MAP_DEPTH_MATRIX K, MAP_DEPTH_MATRIX tau, MAP_DEPTH_MATRIX Lambda,
		BOOLEAN DValueShift) {

// calculate averaging kernel first, because it doesn't need complex conjugate of w
// and calculating conjugate of w changes it

	int kx, ky;
	int Nx, Ny;
	MATRIX wk, wk2PIConj, wkT, Kk, noisek, Lambdak, AvgKernk;
	MATRIX wk2PIConjT, qk, tk;
	MAP_DEPTH_MATRIX w2PIConj;

	Nx = model->w.nX;
	Ny = model->w.nY;

	BOOLEAN shift = DValueShift;
	if (model->q.data == NULL) {
		model->q = newMapDepthMatrix(Nx, Ny, tau.nZ, 1, shift);
	}
	// AvgKern and noise are allocated in LSQRInit; check for null values here before creating new ones.
	if (model->AvgKern.data == NULL) {
		model->AvgKern = newMapDepthMatrixV(Nx, Ny, K.nZ, 1, K.nV, shift);
	}

	if (model->noise.data != NULL) destroyMapDepthMatrix(model->noise);
	shift = TRUE;
	model->noise = newMapDepthMatrix(Nx, Ny, 1, 1, shift);

	// make w*(k)
	w2PIConj = TdmCopy(model->w);
	TdmMultiplyScalarTakeConjugate(&w2PIConj, TWO_PI_SQUARED);

	if (Lambda.invariant) {
		printf("Using invariant Lambda\n");
		Lambdak = extractMapDepth(0, 0, Lambda);
	}

	if (DEBUG1) {
		MATRIX xyplane;
		xyplane = extractXYPlaneFromMapDepthMatrix(0, 0, Lambda);
		if (Lambda.shift)
			printf("Lambda shift is TRUE\n");
		else
			printf("Lambda shift is FALSE\n");
		print_matrix2("Lambda", xyplane, ROW_MAJOR);
		freeMatrix(xyplane);
		if (model->w.shift)
			printf("w shift is TRUE\n");
		else
			printf("w shift is FALSE\n");
		xyplane = extractXYPlaneFromMapDepthMatrix(0, 0, model->w);
		print_matrix2("w", xyplane, ROW_MAJOR);
		freeMatrix(xyplane);
	}
	//for each wavenumber, k = (kx, ky)
	for (kx = 0; kx < Nx; kx++) {
		for (ky = 0; ky < Ny; ky++) {
			// extract w(k)
			wk = extractMapDepth(kx, ky, model->w);
			if (DEBUG) print_matrix2("wk", wk, ROW_MAJOR);

			// extract w*(k)
			wk2PIConj = extractMapDepth(kx, ky, w2PIConj);
			if (DEBUG) print_matrix2("w*(k)", wk2PIConj, ROW_MAJOR);

			// extract Lambda(k)
			if (!Lambda.invariant) {
				Lambdak = extractMapDepth(kx, ky, Lambda);
			}
			if (DEBUG) print_matrix2("Lambdak", Lambdak, ROW_MAJOR);


			// extract tK(k)
			Kk = extractMapDepth(kx, ky, K);
			if (DEBUG) print_matrix2("Kk", Kk, ROW_MAJOR);
			// calculate AvgKern(k)
			wkT = transpose(wk);

			// extract t(k)
			tk = extractMapDepth(kx, ky, tau);
			if (DEBUG) print_matrix2("tk", tk, ROW_MAJOR);

			// calculate q(k)
			// model.q(k) = w*(k) * t(k);
			wk2PIConjT = transpose(wk2PIConj);
			qk = multiplyMM2('N', 'N', wk2PIConjT, tk);
			if (DEBUG) print_matrix2("qk", qk, ROW_MAJOR);
			// insert q(k)
			insertMapDepth(kx, ky, qk, &(model->q));

			// calculate AvgKern
			MATRIX temp = multiplyMM2('N', 'N', wkT, Kk);
			AvgKernk = multiplyScalar(temp, TWO_PI_SQUARED);  // KEEP THIS DEFINITION OF AVG KERN
			freeMatrix(temp);
			if (DEBUG) print_matrix2("AvgKernk", AvgKernk, ROW_MAJOR);
			// insert AvgKernk(k)
			insertMapDepth(kx, ky, AvgKernk, &(model->AvgKern));

			// calculate Lambda(k)
			// noise^2 = w*'(k) Lambda(k) w(k)
			temp = multiplyMM2('N', 'N', wk2PIConjT, Lambdak);
			//print_matrix2("temp", temp, ROW_MAJOR);
			noisek = multiplyMM2('N', 'N', temp, wk);
			freeMatrix(temp);
			temp = noisek;
			noisek = multiplyScalar(temp, TWO_PI_SQUARED);
			freeMatrix(temp);

			// insert noise(k)
			insertMapDepth(kx, ky, noisek, &(model->noise));

			// free some memory
			freeMatrix(wk2PIConjT);
			freeMatrix(qk);
			freeMatrix(AvgKernk);
			freeMatrix(noisek);
			freeMatrix(wkT);
		}
	}
	destroyMapDepthMatrix(w2PIConj);
}

/**
 * Write solution, q, for OLA inversion.
 * @param[in,out] model A Model object.
 * @param[in] baseName BaseDirectory structure for managing output file name.
 */
void writeQ(Model *model, BaseDirectory *baseName) {
	BOOLEAN useFT = TRUE;
	char *filenames[1], *filenameBase;

	BOOLEAN debugPosition = FALSE;
	BOOLEAN compress = FALSE;
	if (IgnoreK0()) {
		filenameBase = Constants.QK0Filename;
	} else {
		filenameBase = Constants.QFilename;
	}
	filenames[0] = finishFilename(baseName, filenameBase, compress, debugPosition);
	writeMapDepthMatrix(filenames, 1, model->q, useFT, model->parms);
}

/**
 * Write OLA inversion weights, w, to file.
 * @param[in,out] model A Model object.
 * @param[in] baseName BaseDirectory structure for managing output file name.
 */
void writeW(Model *model, BaseDirectory *baseName) {
	BOOLEAN useFT = TRUE;
	char *filename, *filenameBase;

	BOOLEAN debugPosition = FALSE;
	BOOLEAN compress = FALSE;
	if (IgnoreK0()) {
		filenameBase = Constants.WK0Filename;
	} else {
		filenameBase = Constants.WFilename;
	}
	filename = finishFilename(baseName, filenameBase, compress, debugPosition);

	printf("writing w to %s\n", filename);
	writeMapDepthMatrixToOneFile(filename, model->w, useFT, model->parms);
}

/**
 * Write averaging kernel to file for OLA inversion.
 * @param[in,out] model A Model object.
 * @param[in] baseName BaseDirectory structure for managing output file name.
 */
void writeAvgKern(Model *model, BaseDirectory *baseName) {
	BOOLEAN useFT = TRUE;
	BOOLEAN debugPosition = FALSE;
	BOOLEAN compress = TRUE;

	int NAlpha;
	NAlpha = model->AvgKern.nV;
	char **filenames, *filenameBase;
	if (IgnoreK0()) {
		filenameBase = Constants.AvgKernK0Filename;
	} else {
		filenameBase = Constants.AvgKernFilename;
	}
	// add NAlpha to the averaging kernel filenames
	filenames = finishFilenames(baseName, filenameBase, NAlpha, compress, debugPosition);
	writeMapDepthMatrix(filenames, NAlpha, model->AvgKern, useFT, model->parms);

}

/**
 * Write the  Averaging Kernel data for 1d problem.
 * @param[in] model The OutputModel containing the Averaging Kernel.
 * @param[in] baseName BaseDirectory structure for managing output file name.
 * @param[in] lambda Which lambda value was used for this solution.
 * @ingroup OutputModel
 */
void writeAVectors1d(Model *model, BaseDirectory *baseName, double lambda) {

		//output A
		if (model->A1d.data == NULL) {
			reportError(NullMatrix,
					"Averaging Kernel (Matrix A) in Model", ERROR);
		} else {
			// write fits file
			fits2d result;
			result = MatrixToFitsObject(model->A1d);
			BOOLEAN debugPosition = FALSE;
			BOOLEAN compress = FALSE;
			char *filename;
			filename = finishFilename(baseName, Constants.AvgKernFilename1d, compress, debugPosition);
			writefits2dWithLambda(filename, result, lambda);
			td_free(result.data, result.xDim * result.yDim * sizeof(double));
		}
}

/**
 * Write noise estimate to file for OLA inversion.
 * @param[in,out] model A Model object.
 * @param[in] baseName BaseDirectory structure for managing output file name.
 */
void writeNoise(Model *model, BaseDirectory *baseName) {
	BOOLEAN useFT = TRUE;
	char *filenames[1], *filenameBase;

	BOOLEAN debugPosition = FALSE;
	BOOLEAN compress = TRUE;

	if (IgnoreK0()) {
		filenameBase = Constants.NoiseK0Filename;
	} else {
		filenameBase = Constants.NoiseFilename;
	}
	filenames[0] = finishFilename(baseName, filenameBase, compress, debugPosition);
	writeMapDepthMatrix(filenames, 1, model->noise, useFT, model->parms);
}

/**
 * Write the Error Covariance data for 1d problem.
 * @param[in] model Model structure containing the Error Covariance.
 * @param[in] baseName Which lambda step was used for this solution.
 * @param[in] lambda Which lambda value was used for this solution.
 * @ingroup OutputModel
 */
void writeCovVectors1d(Model *model, BaseDirectory *baseName, double lambda) {

		//output A
		if (model->Cov1d.data == NULL) {
			reportError(NullMatrix, "Covariance Matrix in Model", ERROR);
		} else {
			// write fits file
			fits2d result;
			result = MatrixToFitsObject(model->Cov1d);

			BOOLEAN debugPosition = FALSE;
			BOOLEAN compress = FALSE;
			char *filename;
			filename = finishFilename(baseName, Constants.CovFilename, compress, debugPosition);
			writefits2dWithLambda(filename, result, lambda);
			td_free(result.data, result.xDim * result.yDim * sizeof(double));
		}
}

/**
 * Write solution matrix for RLS 1D inversion.
 * @param[in,out] model A Model object.
 * @param[in] baseName BaseDirectory structure for managing output file name.
 */
void writeBVectors1d(Model *model, BaseDirectory *baseName) {

	printf("writing solution files ... \n");
	MATRIX bVECTORST = transpose(model->bVECTORS);
	freeMatrix(model->bVECTORS);
	// write B matrix to a fits file

	fits2d result;
	result = MatrixToFitsObject(bVECTORST);

	BOOLEAN debugPosition = FALSE;
	BOOLEAN compress = FALSE;
	char *filename;
	addParmsToFilename2(baseName, NULL, 0, debugPosition);
	filename = finishFilename(baseName, Constants.BFilename, compress, debugPosition);
	writefits2d(filename, result);
	td_free(result.data, result.xDim * result.yDim * sizeof(double));

}

/**
 * Write solution B vectors to FITS files
 * matrix for RLS 3D inversion.
 * @param[in,out] model A Model object.
 * @param[in] baseName BaseDirectory structure for managing output file name.
 */
void writeBVectors3d(Model *model, BaseDirectory *baseName) {
	int useFT = TRUE;
	char **filenames;
	filenames = constructBFilenames(baseName, Constants.BFilename, model->bFull.nV);
	writeMapDepthMatrix(filenames, model->bFull.nT * model->bFull.nV, model->bFull, useFT, model->parms);
}

/**
 * Save lambda values and kx, ky values for each wave number  (RLS 3D inversion).
 * @param[in,out] m A Model object.
 * @param[in] r Row (index into matrix).
 * @param[in] c Column (index into matrix).
 * @param[in] kx Value of kx at (x,y)
 * @param[in] ky Value of ky at (x,y)
 * @param[in] lambda Value of lambda (regularization) parameter.
 */
void saveGridValues(Model *m, int r, int c, double kx, double ky, double lambda) {
	double *p;
	int idx;

	idx = ROW_MAJOR_IDX(r, c, m->kxGridValues.nCol);
	if (!m->kxkyDone) {
		p = m->kxGridValues.data;
		p[idx] = kx; // m.kx;
		p = m->kyGridValues.data;
		p[idx] = ky; //m.ky;
	}
	p = m->lambdaGridValues.data;
	p[idx] = lambda; //  m.lambda;
}

/**
 * Write lambda values and kx, ky values for each wave number  (RLS 3D inversion).
 * Write kxGridValues, kyGridValues and lambdaGridValues at each grid point to FITS files
 * matrix for RLS 3D inversion.
 * @param[in,out] model A Model object.
 * @param[in] baseName BaseDirectory structure for managing output file name.
 */
void writeGridValues3d(Model *model, BaseDirectory *baseName) {
	// write kxGridValues, kyGridValues and lambdaGridValues at each grid point to FITS files
	char* filename;
	BOOLEAN debugPosition = FALSE;
	BOOLEAN compress = FALSE;

	filename = finishFilename(baseName, Constants.LAMBDA_FILENAME, compress, debugPosition);
	writefits2d(filename, MatrixToFitsObject(model->lambdaGridValues));

	if (!model->kxkyDone) {
		addParmsToFilename2(baseName, NULL, 0, debugPosition);
		filename = finishFilename(baseName, Constants.KX_FILENAME, compress, debugPosition);
		writefits2d(filename, MatrixToFitsObject(model->kxGridValues));
		filename = finishFilename(baseName, Constants.KY_FILENAME, compress, debugPosition);
		writefits2d(filename, MatrixToFitsObject(model->kyGridValues));
		model->kxkyDone = TRUE;
	}
}

/**
 * Write d~ matrices for RLS 3D inversion.
 * @param[in,out] model A Model object.
 * @param[in] baseName BaseDirectory structure for managing output file name.
 * @param[in] useFT Boolean indicates if d~ matrices should be Fourier transformed.
 */
void writeDTildaValues3d(Model *model, BaseDirectory *baseName, BOOLEAN useFT) {
	//int useFT = TRUE;
	char **filenames;
	int nFiles;
	int fnameLength;

	printf("writing d~ Files ... \n");
	nFiles = model->dTilde.nT;
	filenames = constructDTildaFilenames(baseName, Constants.DTILDA_FILENAME, nFiles, &fnameLength);
	printf("writing to %s\n", filenames[0]);
	writeMapDepthMatrix(filenames, nFiles, model->dTilde, useFT, model->parms);
	freeNFilenames(filenames, nFiles, fnameLength);
}

/**
 * Write averaging kernels and error covariance matrices for RLS 3D inversion.
 * @param[in,out] model A Model object.
 * @param[in] baseName BaseDirectory structure for managing output file name.
 */
void writeAvgKernValues3d(Model *model, BaseDirectory *baseName) {

	if (calcAvgKernelCov()) {
		int useFT = TRUE;
		char **filenames;
		int nAvgKernels;
		int nMaps;
		int nVectors;
		int nFiles;
		int fnameLength;

		printf("writing Averaging Kernels  ... \n");
		// multiply by (1/(2pi)^2 to correct for ??; bug detected in superg test data
		TdmMultiplyScalar(&(model->AvgKern), 1.0/TWO_PI_SQUARED);

		nMaps = model->AvgKern.nT;
		nVectors = model->AvgKern.nV;

		if (subsetKernels()) {
			nAvgKernels = getNumberOfAveragingKernels();
		} else {
			nAvgKernels = nMaps;
		}

		nFiles = nAvgKernels * nVectors * nVectors;
		filenames = constructAvgCovFilenames(baseName,
				Constants.AvgKernFilename, nFiles, nAvgKernels, nVectors, &fnameLength);

		writeMapDepthMatrix(filenames, nFiles, model->AvgKern, useFT, model->parms);
		freeNFilenames(filenames, nFiles, fnameLength);

		printf("writing Error Covariance Files ... \n");
		// multiply by ((2pi)/N)^2 to correct for two Fourier transforms instead of one
		double factorForCI;
		double x;
		x = model->noise.nX;
		factorForCI = ((2.0*M_PI)*(2.0*M_PI)) / (x * x);
		TdmMultiplyScalar(&(model->noise), factorForCI);

		filenames = constructAvgCovFilenames(baseName,
				Constants.CovFilename, nFiles, nAvgKernels, nVectors, &fnameLength);

		writeMapDepthMatrix(filenames, nFiles, model->noise, useFT, model->parms);
		freeNFilenames(filenames, nFiles, fnameLength);
	}
}


